import logging
from datetime import datetime as dt

import bson
import pymongo

from .bson_store import BSONStore
from .._util import indent
from ..decorators import mongo_retry

logger = logging.getLogger(__name__)

SEQUENCE_STORE_TYPE = "SequenceStore"


class SequenceStore(BSONStore):
    """
    Sequence Store. This stores metadata with timestamps to allow temporal queries.

    Entries are stored in the following format:
        'symbol': symbol name
        'metadata': metadata to be persisted

    """

    @classmethod
    def initialize_library(cls, arctic_lib, hashed=True, **kwargs):
        SequenceStore(arctic_lib)._ensure_index()
        BSONStore.initialize_library(arctic_lib, hashed, **kwargs)

    @mongo_retry
    def _ensure_index(self):
        self.create_index([("symbol", pymongo.ASCENDING)], unique=True, background=True)

    def _reset(self):
        self._collection = self._arctic_lib.get_top_level_collection().metadata

    def __getstate__(self):
        return {"arctic_lib": self._arctic_lib}

    def __setstate__(self, state):
        return SequenceStore.__init__(self, state["arctic_lib"])

    def __str__(self):
        return """<%s at %s>\n%s""" % (self.__class__.__name__, hex(id(self)), indent(str(self._arctic_lib), 4))

    def __repr__(self):
        return str(self)

    @mongo_retry
    def list_symbols(self, regex=None, **kwargs):
        """
        Return the symbols in this library.

        Parameters
        ----------
        regex : `str`
            filter symbols by the passed in regular expression
        kwargs :
            kwarg keys are used as fields to query for symbols with metadata matching
            the kwargs query

        Returns
        -------
        String list of symbols in the library
        """

        # Skip aggregation pipeline
        if not (regex or kwargs):
            return self.distinct("symbol")

        # Index-based query part
        index_query = {}

        if regex:
            # make sure that symbol is present in query even if only as_of is specified to avoid document scans
            # see 'Pipeline Operators and Indexes' at
            # https://docs.mongodb.com/manual/core/aggregation-pipeline/#aggregation-pipeline-operators-and-performance
            index_query["symbol"] = {"$regex": regex or "^"}

        # Document query part
        data_query = {}
        if kwargs:
            for k, v in kwargs.items():
                data_query["metadata." + k] = v

        # Sort using index, relying on https://docs.mongodb.com/manual/core/aggregation-pipeline-optimization/
        pipeline = [{"$sort": {"symbol": pymongo.ASCENDING}}]

        # Index-based filter on symbol and start_time
        if index_query:
            pipeline.append({"$match": index_query})
        # Group by 'symbol' and get the latest known data
        pipeline.append({"$group": {"_id": "$symbol", "metadata": {"$first": "$metadata"}}})
        # Match the data fields
        if data_query:
            pipeline.append({"$match": data_query})
        # Return only 'symbol' field value
        pipeline.append({"$project": {"_id": 0, "symbol": "$_id"}})

        return sorted(r["symbol"] for r in self.aggregate(pipeline))

    @mongo_retry
    def has_symbol(self, symbol):
        return self.find_one({"symbol": symbol}) is not None

    @mongo_retry
    def read(self, symbol):
        """
        Return current metadata saved for `symbol`

        Parameters
        ----------
        symbol : `str`
            symbol name for the item

        Returns
        -------
        metadata
        """
        res = self.find_one({"symbol": symbol})
        return res["metadata"] if res is not None else None

    @mongo_retry
    def get_last_update_time(self, symbol):
        """
        Return last update time of current metadata saved for `symbol`

        Parameters
        ----------
        symbol : `str`
            symbol name for the item

        Returns
        -------
        `datetime.datetime`
        """
        res = self.find_one({"symbol": symbol})
        return res.get("last_update") if res is not None else None

    def write(self, symbol, metadata, last_update=None, overwrite=True):
        """
        Update metadata entry for `symbol`

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        metadata : `dict`
            to be persisted
        last_update : `datetime.datetime`
            last update time
        overwrite: `bool`
            whether to overwrite existing metadata
        """
        if last_update is None:
            last_update = dt.utcnow()
        old_metadata = self.find_one({"symbol": symbol})
        if old_metadata is not None:
            if old_metadata["metadata"] == metadata:
                old_metadata["last_update"] = last_update
                self.find_one_and_update({"symbol": symbol}, {"$set": {"last_update": last_update}})
                return old_metadata
            elif overwrite:
                return self.find_one_and_update(
                    {"symbol": symbol}, {"$set": {"metadata": metadata, "last_update": last_update}}
                )
            else:
                raise ValueError("conflict data")
        elif metadata is None:
            return None

        document = {"_id": bson.ObjectId(), "symbol": symbol, "metadata": metadata, "last_update": last_update}
        mongo_retry(self.insert_one)(document)

        logger.debug("Finished writing metadata for %s", symbol)
        return document

    def append(self, symbol, new_element, last_update=None):
        if last_update is None:
            last_update = dt.utcnow()
        if isinstance(new_element, list):
            return self.find_one_and_update(
                {"symbol": symbol},
                {"$push": {"metadata": {"$each": new_element}}, "$set": {"last_update": last_update}},
            )
        else:
            return self.find_one_and_update(
                {"symbol": symbol}, {"$push": {"metadata": new_element}, "$set": {"last_update": last_update}}
            )

    def remove(self, symbol, value):
        self.update_one({"symbol": symbol}, {"$pull": {"metadata": value}})

    def insert(self, symbol, index, value, last_update=None):
        if last_update is None:
            last_update = dt.utcnow()
        return self.find_one_and_update(
            {"symbol": symbol},
            {"$push": {"metadata": {"$each": [value], "$position": index}}, "$set": {"last_update": last_update}},
        )

    @mongo_retry
    def delete(self, symbol):
        """
        Delete all metadata of `symbol`

        Parameters
        ----------
        symbol : `str`
            symbol name to delete
        """
        self.delete_one({"symbol": symbol})

    def rename(self, from_symbol, to_symbol):
        """

        Args:
            from_symbol:
            to_symbol:

        Returns:

        """
        self.update_many({"symbol": from_symbol}, {"$set": {"symbol": to_symbol}})

    @mongo_retry
    def get_nth_element(self, symbol, index):
        """
        Get nth metadata of `symbol`

        Parameters
        ----------
        symbol : `str`
            symbol name to delete
        n : `int`
            index of metadata

        Returns
        -------
        nth metadata
        """
        pipline = [
            {"$match": {"symbol": symbol}},
            {"$project": {"metadata": {"$arrayElemAt": ["$metadata", index]}}},
        ]
        agg_res = self.aggregate(pipline).next()
        return agg_res.get("metadata", None)

    def get_slice_values(self, symbol: str, start: int, num: int):
        return self.find_one(
            {"symbol": symbol},
            {"metadata": {"$slice": [start, num]}}
        ).get("metadata", [])

    def set_nth_element(self, symbol: str, index: int, value):
        return self.find_one_and_update({"symbol": symbol}, {"$set": {f"metadata.{index}": value}})



import time
import uuid

from enum import Enum

from arctic.exceptions import RequestDurationException


class AsyncRequestType(Enum):
    MODIFIER = 'modifier'
    ACCESSOR = 'accessor'


class AsyncRequest(object):
    def __init__(self, kind, library, fun, callback, *args, **kwargs):
        self.id = uuid.uuid4()

        # Request library call spec
        self.fun = fun
        self.args = args
        self.kwargs = kwargs

        # Request meta
        self.kind = kind
        self.library = library
        self.symbol = kwargs.get('symbol')

        # Request's state
        self.future = None
        self.callback = callback
        self.data = None
        self.exception = None
        self.is_running = False
        self.is_completed = False

        # Timekeeping
        self.start_time = None
        self.end_time = None
        self.create_time = time.time()

        self.mongo_retry = bool(kwargs.get('mongo_retry'))

    @property
    def execution_duration(self):
        if None in (self.start_time, self.end_time):
            raise RequestDurationException("{} can't provide an execution_duration {}.".format(
                self, (self.start_time, self.end_time)))
        return self.end_time - self.start_time

    @property
    def schedule_delay(self):
        if None in (self.start_time, self.create_time):
            raise RequestDurationException("{} can't provide a schedule_delay {}.".format(
                self, (self.start_time, self.create_time)))
        return self.start_time - self.create_time

    @property
    def total_time(self):
        if None in (self.end_time, self.create_time):
            raise RequestDurationException("{} can't provide a total_time {}.".format(
                self, (self.end_time, self.create_time)))
        return self.end_time - self.create_time

    def __str__(self):
        return "Request id:{} library:{}, symbol:{} fun:{}, kind:{}".format(
            self.id, self.library, self.symbol, getattr(self.fun, '__name__', None), self.kind)



import random
from datetime import datetime as dt
from datetime import timedelta

from pandas import DataFrame, Index, MultiIndex
from pandas.testing import assert_frame_equal
from tests.util import assert_frame_equal_

from arctic.chunkstore.utils import read_apply


def create_test_data(size=5, index=True, multiindex=True, random_data=True, random_ids=True, date_offset=0, use_hours=False, cols=1):
    data = {}
    for i in range(cols):
        if random_data:
            data['data' + str(i)] = [random.random() * random.randint(-100, 100) for _ in range(size)]
        else:
            data['data' + str(i)] = range(size)
    dates = [dt(2016, 1, 1) + timedelta(days=0 if use_hours else n+date_offset,
                                        hours=n+date_offset if use_hours else 0) for n in range(size)]
    if index:
        if multiindex:
            index_col_names = ['date', 'id']
            idx = [(date, random.randint(1, size)) for date in dates] if random_ids else [(date, 1) for date in dates]
            index = MultiIndex.from_tuples(idx, names=index_col_names) if idx else MultiIndex([[]]*2, [[]]*2, names=index_col_names)
            return DataFrame(data=data, index=index)
        return DataFrame(data=data, index=Index(data=dates, name='date'))
    data.update({'date': dates})
    return DataFrame(data=data)


def test_read_apply(chunkstore_lib):
    df = create_test_data(index=False, size=20)
    chunkstore_lib.write('test', df, chunk_size='M')

    def func(df):
        df['data0'] += 1.0
        return df

    for data in read_apply(chunkstore_lib, 'test', func):
        assert_frame_equal_(data, func(df))



import logging
from datetime import datetime as dt, timedelta

import bson
import pymongo
from pymongo import ReadPreference
from pymongo.errors import OperationFailure, AutoReconnect, DuplicateKeyError

from ._pickle_store import PickleStore
from ._version_store_utils import cleanup, get_symbol_alive_shas, _get_symbol_pointer_cfgs
from .versioned_item import VersionedItem
from .._config import STRICT_WRITE_HANDLER_MATCH, FW_POINTERS_REFS_KEY, FW_POINTERS_CONFIG_KEY, FwPointersCfg
from .._util import indent, enable_sharding, mongo_count, get_fwptr_config
from ..date import mktz, datetime_to_ms, ms_to_datetime
from ..decorators import mongo_retry
from ..exceptions import NoDataFoundException, DuplicateSnapshotException, \
    ArcticException
from ..hooks import log_exception

logger = logging.getLogger(__name__)

VERSION_STORE_TYPE = 'VersionStore'
_TYPE_HANDLERS = []
ARCTIC_VERSION = None
ARCTIC_VERSION_NUMERICAL = None


def register_version(version, numerical):
    global ARCTIC_VERSION, ARCTIC_VERSION_NUMERICAL
    ARCTIC_VERSION = version
    ARCTIC_VERSION_NUMERICAL = numerical


def register_versioned_storage(storageClass):
    existing_instances = [i for i, v in enumerate(_TYPE_HANDLERS) if str(v.__class__) == str(storageClass)]
    if existing_instances:
        for i in existing_instances:
            _TYPE_HANDLERS[i] = storageClass()
    else:
        _TYPE_HANDLERS.append(storageClass())
    return storageClass


class VersionStore(object):

    _bson_handler = PickleStore()

    @classmethod
    def initialize_library(cls, arctic_lib, hashed=True, **kwargs):
        c = arctic_lib.get_top_level_collection()

        if 'strict_write_handler' in kwargs:
            arctic_lib.set_library_metadata('STRICT_WRITE_HANDLER_MATCH',
                                            bool(kwargs.pop('strict_write_handler')))

        for th in _TYPE_HANDLERS:
            th.initialize_library(arctic_lib, **kwargs)
        VersionStore._bson_handler.initialize_library(arctic_lib, **kwargs)
        VersionStore(arctic_lib)._ensure_index()

        logger.info("Trying to enable sharding...")
        try:
            enable_sharding(arctic_lib.arctic, arctic_lib.get_name(), hashed=hashed)
        except OperationFailure as e:
            logger.warning("Library created, but couldn't enable sharding: %s. This is OK if you're not 'admin'" % str(e))

    @mongo_retry
    def _last_version_seqnum(self, symbol):
        last_seq = self._version_nums.find_one({'symbol': symbol})
        return last_seq['version'] if last_seq else 0

    # new index named because of 127 char restriction on fully qualified index names
    @mongo_retry
    def _ensure_index(self):
        collection = self._collection
        collection.snapshots.create_index([('name', pymongo.ASCENDING)], unique=True,
                                          background=True)
        collection.versions.create_index([('symbol', pymongo.ASCENDING), ('_id', pymongo.DESCENDING)],
                                         background=True)
        collection.versions.create_index([('symbol', pymongo.ASCENDING), ('version', pymongo.DESCENDING)], unique=True,
                                         background=True)
        collection.versions.create_index([('symbol', pymongo.ASCENDING), ('version', pymongo.DESCENDING),
                                          ('metadata.deleted', pymongo.ASCENDING)],
                                         name='versionstore_idx',
                                         background=True)
        # Issue #987 slow snapshot delete
        collection.versions.create_index([('parent', pymongo.ASCENDING)], background=True)
        collection.version_nums.create_index('symbol', unique=True, background=True)
        for th in _TYPE_HANDLERS:
            th._ensure_index(collection)

    def __init__(self, arctic_lib):
        self._arctic_lib = arctic_lib
        # Do we allow reading from secondaries
        self._allow_secondary = self._arctic_lib.arctic._allow_secondary
        self._reset()
        self._with_strict_handler = None

    @property
    def _with_strict_handler_match(self):
        if self._with_strict_handler is None:
            strict_meta = self._arctic_lib.get_library_metadata('STRICT_WRITE_HANDLER_MATCH')
            self._with_strict_handler = STRICT_WRITE_HANDLER_MATCH if strict_meta is None else strict_meta
        return self._with_strict_handler

    @mongo_retry
    def _reset(self):
        # The default collections
        self._collection = self._arctic_lib.get_top_level_collection()
        self._audit = self._collection.audit
        self._snapshots = self._collection.snapshots
        self._versions = self._collection.versions
        self._version_nums = self._collection.version_nums

    def __getstate__(self):
        return {'arctic_lib': self._arctic_lib}

    def __setstate__(self, state):
        return VersionStore.__init__(self, state['arctic_lib'])

    def __str__(self):
        return """<%s at %s>
%s""" % (self.__class__.__name__, hex(id(self)), indent(str(self._arctic_lib), 4))

    def __repr__(self):
        return str(self)

    def _read_preference(self, allow_secondary):
        """ Return the mongo read preference given an 'allow_secondary' argument
        """
        allow_secondary = self._allow_secondary if allow_secondary is None else allow_secondary
        return ReadPreference.NEAREST if allow_secondary else ReadPreference.PRIMARY

    @mongo_retry
    def list_symbols(self, all_symbols=False, snapshot=None, regex=None, **kwargs):
        """
        Return the symbols in this library.

        Parameters
        ----------
        all_symbols : `bool`
            If True returns all symbols under all snapshots, even if the symbol has been deleted
            in the current version (i.e. it exists under a snapshot... Default: False
        snapshot : `str`
            Return the symbols available under the snapshot.
        regex : `str`
            filter symbols by the passed in regular expression
        kwargs :
            kwarg keys are used as fields to query for symbols with metadata matching
            the kwargs query

        Returns
        -------
        String list of symbols in the library
        """
        query = {}
        if regex is not None:
            query['symbol'] = {'$regex': regex}
        if kwargs:
            for k, v in kwargs.items():
                # TODO: this doesn't work as expected as it ignores the versions with metadata.deleted set
                #       as a result it will return symbols with matching metadata which have been deleted
                #       Maybe better add a match step in the pipeline instead of making it part of the query
                query['metadata.' + k] = v
        if snapshot is not None:
            try:
                query['parent'] = self._snapshots.find_one({'name': snapshot})['_id']
            except TypeError:
                raise NoDataFoundException('No snapshot %s in library %s' % (snapshot, self._arctic_lib.get_name()))
        elif all_symbols:
            return self._versions.find(query).distinct('symbol')

        # Return just the symbols which aren't deleted in the 'trunk' of this library
        pipeline = []
        if query:
            # Match based on user criteria first
            pipeline.append({'$match': query})
        pipeline.extend([
            {'$sort': bson.SON([('symbol', pymongo.ASCENDING), ('version', pymongo.DESCENDING)])},
            {'$group': {
                '_id': '$symbol',
                'deleted': {'$first': '$metadata.deleted'}
            }},
            {'$match': {'deleted': {'$ne': True}}}
        ])
        # We may hit the group memory limit (100MB), so use allowDiskUse to circumvent this
        #  - https://docs.mongodb.com/manual/reference/operator/aggregation/group/#group-memory-limit
        return sorted([x['_id'] for x in self._versions.aggregate(pipeline, allowDiskUse=True)])

    @mongo_retry
    def has_symbol(self, symbol, as_of=None):
        """
        Return True if the 'symbol' exists in this library AND the symbol
        isn't deleted in the specified as_of.

        It's possible for a deleted symbol to exist in older snapshots.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        as_of : `str` or int or `datetime.datetime`
            Return the data as it was as_of the point in time.
            `int` : specific version number
            `str` : snapshot name which contains the version
            `datetime.datetime` : the version of the data that existed as_of the requested point in time
        """
        try:
            # Always use the primary for has_symbol, it's safer
            self._read_metadata(symbol, as_of=as_of, read_preference=ReadPreference.PRIMARY)
            return True
        except NoDataFoundException:
            return False

    def read_audit_log(self, symbol=None, message=None):
        """
        Return the audit log associated with a given symbol

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        """
        query = {}
        if symbol:
            if isinstance(symbol, str):
                query['symbol'] = {'$regex': symbol}
            else:
                query['symbol'] = {'$in': list(symbol)}

        if message is not None:
            query['message'] = message

        def _pop_id(x):
            x.pop('_id')
            return x

        return [_pop_id(x) for x in self._audit.find(query, sort=[('_id', -1)])]

    def list_versions(self, symbol=None, snapshot=None, latest_only=False):
        """
        Return a list of versions filtered by the passed in parameters.

        Parameters
        ----------
        symbol : `str`
            Symbol to return versions for.  If None returns versions across all
            symbols in the library.
        snapshot : `str`
            Return the versions contained in the named snapshot
        latest_only : `bool`
            Only include the latest version for a specific symbol

        Returns
        -------
        List of dictionaries describing the discovered versions in the library
        """
        if symbol is None:
            symbols = self.list_symbols(snapshot=snapshot)
        else:
            symbols = [symbol]

        query = {}

        if snapshot is not None:
            try:
                query['parent'] = self._snapshots.find_one({'name': snapshot})['_id']
            except TypeError:
                raise NoDataFoundException('No snapshot %s in library %s' % (snapshot, self._arctic_lib.get_name()))

        versions = []
        snapshots = {ss.get('_id'): ss.get('name') for ss in self._snapshots.find()}
        for symbol in symbols:
            query['symbol'] = symbol
            seen_symbols = set()
            for version in self._versions.find(query, projection=['symbol', 'version', 'parent', 'metadata.deleted'], sort=[('version', -1)]):
                if latest_only and version['symbol'] in seen_symbols:
                    continue
                seen_symbols.add(version['symbol'])
                meta = version.get('metadata')
                versions.append({'symbol': version['symbol'], 'version': version['version'],
                                 'deleted': meta.get('deleted', False) if meta else False,
                                 # We return offset-aware datetimes in Local Time.
                                 'date': ms_to_datetime(datetime_to_ms(version['_id'].generation_time)),
                                 'snapshots': [snapshots[s] for s in version.get('parent', []) if s in snapshots]})
        return versions

    def _find_snapshots(self, parent_ids):
        snapshots = []
        for p in parent_ids:
            snap = self._snapshots.find_one({'_id': p})
            if snap:
                snapshots.append(snap['name'])
            else:
                snapshots.append(str(p))
        return snapshots

    def _read_handler(self, version, symbol):
        handler = None
        for h in _TYPE_HANDLERS:
            if h.can_read(version, symbol):
                handler = h
                break
        if handler is None:
            handler = self._bson_handler
        return handler

    @staticmethod
    def handler_can_write_type(handler, data):
        type_method = getattr(handler, "can_write_type", None)
        if callable(type_method):
            return type_method(data)
        return False

    def _write_handler(self, version, symbol, data, **kwargs):
        handler = None
        for h in _TYPE_HANDLERS:
            if h.can_write(version, symbol, data, **kwargs):
                handler = h
                break
            if self._with_strict_handler_match and self.handler_can_write_type(h, data):
                raise ArcticException("Not falling back to default handler for %s" % symbol)
        if handler is None:
            version['type'] = 'default'
            handler = self._bson_handler
        return handler

    def read(self, symbol, as_of=None, date_range=None, from_version=None, allow_secondary=None, **kwargs):
        """
        Read data for the named symbol.  Returns a VersionedItem object with
        a data and metdata element (as passed into write).

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        as_of : `str` or `int` or `datetime.datetime`
            Return the data as it was as_of the point in time.
            `int` : specific version number
            `str` : snapshot name which contains the version
            `datetime.datetime` : the version of the data that existed as_of the requested point in time
        date_range: `arctic.date.DateRange`
            DateRange to read data for.  Applies to Pandas data, with a DateTime index
            returns only the part of the data that falls in the DateRange.
        allow_secondary : `bool` or `None`
            Override the default behavior for allowing reads from secondary members of a cluster:
            `None` : use the settings from the top-level `Arctic` object used to query this version store.
            `True` : allow reads from secondary members
            `False` : only allow reads from primary members

        Returns
        -------
        VersionedItem namedtuple which contains a .data and .metadata element
        """
        try:
            read_preference = self._read_preference(allow_secondary)
            _version = self._read_metadata(symbol, as_of=as_of, read_preference=read_preference)
            return self._do_read(symbol, _version, from_version,
                                 date_range=date_range, read_preference=read_preference, **kwargs)
        except (OperationFailure, AutoReconnect) as e:
            # Log the exception so we know how often this is happening
            log_exception('read', e, 1)
            # If we've failed to read from the secondary, then it's possible the
            # secondary has lagged.  In this case direct the query to the primary.
            _version = mongo_retry(self._read_metadata)(symbol, as_of=as_of,
                                                        read_preference=ReadPreference.PRIMARY)
            return self._do_read_retry(symbol, _version, from_version,
                                       date_range=date_range,
                                       read_preference=ReadPreference.PRIMARY,
                                       **kwargs)
        except Exception as e:
            log_exception('read', e, 1)
            raise

    @mongo_retry
    def get_info(self, symbol, as_of=None):
        """
        Reads and returns information about the data stored for symbol

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        as_of : `str` or int or `datetime.datetime`
            Return the data as it was as_of the point in time.
            `int` : specific version number
            `str` : snapshot name which contains the version
            `datetime.datetime` : the version of the data that existed as_of the requested point in time

        Returns
        -------
        dictionary of the information (specific to the type of data)
        """
        version = self._read_metadata(symbol, as_of=as_of, read_preference=None)
        handler = self._read_handler(version, symbol)
        if handler and hasattr(handler, 'get_info'):
            return handler.get_info(version)
        return {}

    @staticmethod
    def handler_supports_read_option(handler, option):
        options_method = getattr(handler, "read_options", None)
        if callable(options_method):
            return option in options_method()

        # If the handler doesn't support interrogation of its read options assume
        # that it does support this option (i.e. fail-open)
        return True

    def get_arctic_version(self, symbol, as_of=None):
        """
        Return the numerical representation of the arctic version used to write the last (or as_of) version for
        the given symbol.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        as_of : `str` or int or `datetime.datetime`
            Return the data as it was as_of the point in time.
            `int` : specific version number
            `str` : snapshot name which contains the version
            `datetime.datetime` : the version of the data that existed as_of the requested point in time

        Returns
        -------
        arctic_version : int
            The numerical representation of Arctic version, used to create the specified symbol version
        """
        return self._read_metadata(symbol, as_of=as_of).get('arctic_version', 0)

    def _do_read(self, symbol, version, from_version=None, **kwargs):
        if version.get('deleted'):
            raise NoDataFoundException("No data found for %s in library %s" % (symbol, self._arctic_lib.get_name()))
        handler = self._read_handler(version, symbol)
        # We don't push the date_range check in the handler's code, since the "_with_strict_handler_match"
        #    value is configured on a per-library basis, and is part of the VersionStore instance.
        if self._with_strict_handler_match and \
                kwargs.get('date_range') and \
                not self.handler_supports_read_option(handler, 'date_range'):
            raise ArcticException("Date range arguments not supported by handler in %s" % symbol)

        data = handler.read(self._arctic_lib, version, symbol, from_version=from_version, **kwargs)
        return VersionedItem(symbol=symbol, library=self._arctic_lib.get_name(), version=version['version'],
                             metadata=version.pop('metadata', None), data=data,
                             host=self._arctic_lib.arctic.mongo_host)
    _do_read_retry = mongo_retry(_do_read)

    @mongo_retry
    def read_metadata(self, symbol, as_of=None, allow_secondary=None):
        """
        Return the metadata saved for a symbol.  This method is fast as it doesn't
        actually load the data.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        as_of : `str` or int or `datetime.datetime`
            Return the data as it was as_of the point in time.
            `int` : specific version number
            `str` : snapshot name which contains the version
            `datetime.datetime` : the version of the data that existed as_of the requested point in time
        allow_secondary : `bool` or `None`
            Override the default behavior for allowing reads from secondary members of a cluster:
            `None` : use the settings from the top-level `Arctic` object used to query this version store.
            `True` : allow reads from secondary members
            `False` : only allow reads from primary members
        """
        _version = self._read_metadata(symbol, as_of=as_of, read_preference=self._read_preference(allow_secondary))
        return VersionedItem(symbol=symbol, library=self._arctic_lib.get_name(), version=_version['version'],
                             metadata=_version.pop('metadata', None), data=None,
                             host=self._arctic_lib.arctic.mongo_host)

    def _read_metadata(self, symbol, as_of=None, read_preference=None):
        if read_preference is None:
            # We want to hit the PRIMARY if querying secondaries is disabled.  If we're allowed to query secondaries,
            # then we want to hit the secondary for metadata.  We maintain ordering of chunks vs. metadata, such that
            # if metadata is available, we guarantee that chunks will be available. (Within a 10 minute window.)
            read_preference = ReadPreference.PRIMARY_PREFERRED if not self._allow_secondary else ReadPreference.SECONDARY_PREFERRED

        versions_coll = self._versions.with_options(read_preference=read_preference)

        _version = None
        if as_of is None:
            _version = versions_coll.find_one({'symbol': symbol}, sort=[('version', pymongo.DESCENDING)])
        elif isinstance(as_of, str):
            # as_of is a snapshot
            snapshot = self._snapshots.find_one({'name': as_of})
            if snapshot:
                _version = versions_coll.find_one({'symbol': symbol, 'parent': snapshot['_id']})
        elif isinstance(as_of, dt):
            # as_of refers to a datetime
            if not as_of.tzinfo:
                as_of = as_of.replace(tzinfo=mktz())
            _version = versions_coll.find_one({'symbol': symbol,
                                               '_id': {'$lt': bson.ObjectId.from_datetime(as_of + timedelta(seconds=1))}},
                                              sort=[('symbol', pymongo.DESCENDING), ('version', pymongo.DESCENDING)])
        else:
            # Backward compatibility - as of is a version number
            _version = versions_coll.find_one({'symbol': symbol, 'version': as_of})

        if not _version:
            raise NoDataFoundException("No data found for %s in library %s" % (symbol, self._arctic_lib.get_name()))

        # if the item has been deleted, don't return any metadata
        metadata = _version.get('metadata', None)
        if metadata is not None and metadata.get('deleted', False) is True:
            raise NoDataFoundException("No data found for %s in library %s" % (symbol, self._arctic_lib.get_name()))

        return _version

    def _insert_version(self, version):
        try:
            # Keep here the mongo_retry to avoid incrementing versions and polluting the DB with garbage segments,
            # upon intermittent Mongo errors
            # If, however, we get a DuplicateKeyError, suppress it and raise OperationFailure, so that the method-scoped
            # mongo_retry re-tries and creates a new version, to overcome the issue.
            mongo_retry(self._versions.insert_one)(version)
        except DuplicateKeyError as err:
            logger.exception(err)
            raise OperationFailure("A version with the same _id exists, force a clean retry")

    @mongo_retry
    def append(self, symbol, data, metadata=None, prune_previous_version=True, upsert=True, **kwargs):
        """
        Append 'data' under the specified 'symbol' name to this library.
        The exact meaning of 'append' is left up to the underlying store implementation.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        data :
            to be persisted
        metadata : `dict`
            an optional dictionary of metadata to persist along with the symbol.
        prune_previous_version : `bool`
            Removes previous (non-snapshotted) versions from the database.
            Default: True
        upsert : `bool`
            Write 'data' if no previous version exists.
        """
        self._arctic_lib.check_quota()
        version = {'_id': bson.ObjectId()}
        version['arctic_version'] = ARCTIC_VERSION_NUMERICAL
        version['symbol'] = symbol
        spec = {'symbol': symbol}
        previous_version = self._versions.find_one(spec,
                                                   sort=[('version', pymongo.DESCENDING)])

        if len(data) == 0 and previous_version is not None:
            return VersionedItem(symbol=symbol, library=self._arctic_lib.get_name(), version=previous_version['version'],
                                 metadata=version.pop('metadata', None), data=None,
                                 host=self._arctic_lib.arctic.mongo_host)

        if upsert and previous_version is None:
            return self.write(symbol=symbol, data=data, prune_previous_version=prune_previous_version, metadata=metadata)

        assert previous_version is not None
        dirty_append = False

        # Take a version number for this append.
        # If the version numbers of this and the previous version aren't sequential,
        # then we've either had a failed append in the past,
        # we're in the midst of a concurrent update, we've deleted a version in-between
        # or are somehow 'forking' history
        next_ver = self._version_nums.find_one_and_update({'symbol': symbol, },
                                                          {'$inc': {'version': 1}},
                                                          upsert=False, new=True)['version']

        # This is a very important check, do not remove/modify as it is a guard-dog preventing potential data corruption
        if next_ver != previous_version['version'] + 1:
            dirty_append = True
            logger.debug('''version_nums is out of sync with previous version document.
            This probably means that either a version document write has previously failed, or the previous version has been deleted.''')

        # if the symbol has previously been deleted then overwrite
        previous_metadata = previous_version.get('metadata', None)
        if upsert and previous_metadata is not None and previous_metadata.get('deleted', False) is True:
            return self.write(symbol=symbol, data=data, prune_previous_version=prune_previous_version,
                              metadata=metadata)

        handler = self._read_handler(previous_version, symbol)

        if metadata is not None:
            version['metadata'] = metadata
        elif 'metadata' in previous_version:
            version['metadata'] = previous_version['metadata']

        if handler and hasattr(handler, 'append') and callable(handler.append):
            handler.append(self._arctic_lib, version, symbol, data,
                           previous_version, dirty_append=dirty_append, **kwargs)
        else:
            raise Exception("Append not implemented for handler %s" % handler)

        if prune_previous_version and previous_version:
            # Does not allow prune to remove the base of the new version
            self._prune_previous_versions(
                symbol,
                keep_version=version.get('base_version_id'),
                new_version_shas=version.get(FW_POINTERS_REFS_KEY),
                keep_mins=kwargs.get('keep_mins', 120)
            )

        # Insert the new version into the version DB
        version['version'] = next_ver
        self._insert_version(version)

        return VersionedItem(symbol=symbol, library=self._arctic_lib.get_name(), version=version['version'],
                             metadata=version.pop('metadata', None), data=None,
                             host=self._arctic_lib.arctic.mongo_host)

    @mongo_retry
    def write(self, symbol, data, metadata=None, prune_previous_version=True, **kwargs):
        """
        Write 'data' under the specified 'symbol' name to this library.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        data :
            to be persisted
        metadata : `dict`
            an optional dictionary of metadata to persist along with the symbol.
            Default: None
        prune_previous_version : `bool`
            Removes previous (non-snapshotted) versions from the database.
            Default: True
        kwargs :
            passed through to the write handler

        Returns
        -------
        VersionedItem named tuple containing the metadata and version number
        of the written symbol in the store.
        """
        self._arctic_lib.check_quota()
        version = {'_id': bson.ObjectId()}
        version['arctic_version'] = ARCTIC_VERSION_NUMERICAL
        version['symbol'] = symbol
        version['version'] = self._version_nums.find_one_and_update({'symbol': symbol},
                                                                    {'$inc': {'version': 1}},
                                                                    upsert=True, new=True)['version']
        version['metadata'] = metadata

        previous_version = self._versions.find_one({'symbol': symbol, 'version': {'$lt': version['version']}},
                                                   sort=[('version', pymongo.DESCENDING)])

        handler = self._write_handler(version, symbol, data, **kwargs)
        handler.write(self._arctic_lib, version, symbol, data, previous_version, **kwargs)

        if prune_previous_version and previous_version:
            self._prune_previous_versions(
                symbol,
                keep_mins=kwargs.get('keep_mins', 120),
                new_version_shas=version.get(FW_POINTERS_REFS_KEY)
            )

        # Insert the new version into the version DB
        self._insert_version(version)

        logger.debug('Finished writing versions for %s', symbol)

        return VersionedItem(symbol=symbol, library=self._arctic_lib.get_name(), version=version['version'],
                             metadata=version.pop('metadata', None), data=None,
                             host=self._arctic_lib.arctic.mongo_host)

    def _add_new_version_using_reference(self, symbol, new_version, reference_version, prune_previous_version):
        # Attention: better not use this method following an append.
        # It is dangerous because if it deletes the version at the last_look, the segments added by the
        # append are dangling (if prune_previous_version is False) and can cause potentially corruption.
        constraints = new_version and \
                      reference_version and \
                      new_version['symbol'] == reference_version['symbol'] and \
                      new_version['_id'] != reference_version['_id'] and \
                      new_version['base_version_id']
        assert constraints
        # There is always a small risk here another process in between these two calls (above/below)
        # to delete the reference_version, which may happen to be the last parent entry in the data segments.
        # In this case the segments will be deleted by the other process,
        # and the new version's "base_version_id" won't be referenced by any segments.
        # Do a naive check for concurrent mods
        lastv_seqn = self._last_version_seqnum(symbol)
        if lastv_seqn != new_version['version']:
            raise OperationFailure("The symbol {} has been modified concurrently ({} != {})".format(
                symbol, lastv_seqn, new_version['version']))

        # Insert the new version into the version DB
        # (must come before the pruning, otherwise base version won't be preserved)
        self._insert_version(new_version)

        # Check if in the meanwhile the reference version (based on which we updated incrementally) has been removed
        last_look = self._versions.find_one({'_id': reference_version['_id']})
        if last_look is None or last_look.get('deleted'):
            # Revert the change
            mongo_retry(self._versions.delete_one)({'_id': new_version['_id']})
            # Indicate the failure
            raise OperationFailure("Failed to write metadata for symbol %s. "
                                   "The previous version (%s, %d) has been removed during the update" %
                                   (symbol, str(reference_version['_id']), reference_version['version']))

        if prune_previous_version and reference_version:
            self._prune_previous_versions(symbol, new_version_shas=new_version.get(FW_POINTERS_REFS_KEY))

        logger.debug('Finished updating versions with new metadata for %s', symbol)

        return VersionedItem(symbol=symbol, library=self._arctic_lib.get_name(), version=new_version['version'],
                             metadata=new_version.get('metadata'), data=None,
                             host=self._arctic_lib.arctic.mongo_host)

    @mongo_retry
    def write_metadata(self, symbol, metadata, prune_previous_version=True, **kwargs):
        """
        Write 'metadata' under the specified 'symbol' name to this library.
        The data will remain unchanged. A new version will be created.
        If the symbol is missing, it causes a write with empty data (None, pickled, can't append)
        and the supplied metadata.
        Returns a VersionedItem object only with a metadata element.
        Fast operation: Zero data/segment read/write operations.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        metadata : `dict` or `None`
            dictionary of metadata to persist along with the symbol
        prune_previous_version : `bool`
            Removes previous (non-snapshotted) versions from the database.
            Default: True
        kwargs :
            passed through to the write handler (only used if symbol does not already exist or is deleted)

        Returns
        -------
        `VersionedItem`
            VersionedItem named tuple containing the metadata of the written symbol's version document in the store.
        """
        # Make a normal write with empty data and supplied metadata if symbol does not exist
        try:
            previous_version = self._read_metadata(symbol)
        except NoDataFoundException:
            return self.write(symbol, data=None, metadata=metadata,
                              prune_previous_version=prune_previous_version, **kwargs)

        # Reaching here means that and/or metadata exist and we are set to update the metadata
        new_version_num = self._version_nums.find_one_and_update({'symbol': symbol},
                                                                 {'$inc': {'version': 1}},
                                                                 upsert=True, new=True)['version']

        # Populate the new version entry, preserving existing data, and updating with the supplied metadata
        version = {k: previous_version[k] for k in previous_version.keys() if k != 'parent'}   # don't copy snapshots
        version['_id'] = bson.ObjectId()
        version['version'] = new_version_num
        version['metadata'] = metadata
        version['base_version_id'] = previous_version.get('base_version_id', previous_version['_id'])

        return self._add_new_version_using_reference(symbol, version, previous_version, prune_previous_version)

    @mongo_retry
    def restore_version(self, symbol, as_of, prune_previous_version=True):
        """
        Restore the specified 'symbol' data and metadata to the state of a given version/snapshot/date.
        Returns a VersionedItem object only with a metadata element.
        Fast operation: Zero data/segment read/write operations.

        Parameters
        ----------
        symbol : `str`
            symbol name for the item
        as_of : `str` or `int` or `datetime.datetime`
            Return the data as it was as_of the point in time.
            `int` : specific version number
            `str` : snapshot name which contains the version
            `datetime.datetime` : the version of the data that existed as_of the requested point in time
        prune_previous_version : `bool`
            Removes previous (non-snapshotted) versions from the database.
            Default: True

        Returns
        -------
        `VersionedItem`
            VersionedItem named tuple containing the metadata of the written symbol's version document in the store.
        """
        # TODO: This operation is tricky as it may create history branches and lead to corrupted symbols.
        #       To avoid this we do concat_rewrite (see Issue #579)
        #       Investigate how this can be optimized and maintain safety (i.e. avoid read/write with serialization
        #       and compression costs, but instead:
        #       clone segments (server-side?) / crate new (base) version / update segments' parent).

        version_to_restore = self._read_metadata(symbol, as_of=as_of)

        # At this point it is guaranteed that the as_of version exists and doesn't have the symbol marked as deleted.
        # If we try to restore the last version, do nothing (No-Op) and return the associated VesionedItem.
        if self._last_version_seqnum(symbol) == version_to_restore['version']:
            return VersionedItem(symbol=symbol, library=self._arctic_lib.get_name(),
                                 version=version_to_restore['version'],
                                 host=self._arctic_lib.arctic.mongo_host,
                                 metadata=version_to_restore.pop('metadata', None), data=None)

        # Read the existing data from as_of
        item = self.read(symbol, as_of=as_of)

        # Write back, creating a new base version
        new_item = self.write(symbol,
                              data=item.data, metadata=item.metadata, prune_previous_version=prune_previous_version)
        return new_item

    @mongo_retry
    def _find_prunable_version_ids(self, symbol, keep_mins):
        """
        Find all non-snapshotted versions of a symbol that are older than a version that's at least keep_mins
        minutes old.

        Based on documents available on the secondary.
        """
        read_preference = ReadPreference.SECONDARY_PREFERRED if keep_mins > 0 else ReadPreference.PRIMARY
        versions = self._versions.with_options(read_preference=read_preference)
        query = {'symbol': symbol,
                 # Not snapshotted
                 '$or': [{'parent': {'$exists': False}}, {'parent': []}],
                 # At least 'keep_mins' old
                 '_id': {'$lt': bson.ObjectId.from_datetime(dt.utcnow()
                                                            # Add one second as the ObjectId
                                                            # str has random fuzz
                                                            + timedelta(seconds=1)
                                                            - timedelta(minutes=keep_mins)
                                                            )
                         }
                 }
        cursor = versions.find(query,
                               # Using version number here instead of _id as there's a very unlikely case
                               # where the versions are created on different hosts or processes at exactly
                               # the same time.
                               sort=[('version', pymongo.DESCENDING)],
                               # Guarantees at least one version is kept
                               skip=1,
                               projection={'_id': 1, FW_POINTERS_REFS_KEY: 1, FW_POINTERS_CONFIG_KEY: 1},
                               )
        return {v['_id']: ([bson.binary.Binary(x) for x in v.get(FW_POINTERS_REFS_KEY, [])], get_fwptr_config(v))
                for v in cursor}

    @mongo_retry
    def _find_base_version_ids(self, symbol, version_ids):
        """
        Return all base_version_ids for a symbol that are not bases of version_ids
        """
        cursor = self._versions.find({'symbol': symbol,
                                      '_id': {'$nin': version_ids},
                                      'base_version_id': {'$exists': True},
                                      },
                                     projection={'base_version_id': 1})
        return [version["base_version_id"] for version in cursor]

    def _prune_previous_versions(self, symbol, keep_mins=120, keep_version=None, new_version_shas=None):
        """
        Prune versions, not pointed at by snapshots which are at least keep_mins old. Prune will never
        remove all versions.
        """
        new_version_shas = new_version_shas if new_version_shas else []
        prunable_ids_to_shas = self._find_prunable_version_ids(symbol, keep_mins)
        prunable_ids = list(prunable_ids_to_shas.keys())
        if keep_version is not None:
            try:
                prunable_ids.remove(keep_version)
            except ValueError:
                pass
        if not prunable_ids:
            return

        base_version_ids = self._find_base_version_ids(symbol, prunable_ids)
        version_ids = list(set(prunable_ids) - set(base_version_ids))
        if not version_ids:
            return

        # Delete the version documents
        mongo_retry(self._versions.delete_many)({'_id': {'$in': version_ids}})

        prunable_ids_to_shas = {k: prunable_ids_to_shas[k] for k in version_ids}

        # The new version has not been written yet, so make sure that any SHAs pointed by it are preserved
        shas_to_delete = [sha for v in prunable_ids_to_shas.values() for sha in v[0] if sha not in new_version_shas]

        # Cleanup any chunks
        mongo_retry(cleanup)(self._arctic_lib, symbol, version_ids, self._versions,
                             shas_to_delete=shas_to_delete,
                             pointers_cfgs=[v[1] for v in prunable_ids_to_shas.values()])

    @mongo_retry
    def _delete_version(self, symbol, version_num, do_cleanup=True):
        """
        Delete the n'th version of this symbol from the historical collection.
        """
        version = self._versions.find_one({'symbol': symbol, 'version': version_num})
        if not version:
            logger.error("Can't delete %s:%s as not found in DB" % (symbol, version_num))
            return
        # If the version is pointed to by a snapshot, then can't delete
        if version.get('parent', None):
            for parent in version['parent']:
                snap_name = self._snapshots.find_one({'_id': parent})
                if snap_name:
                    snap_name = snap_name['name']
                logger.error("Can't delete: %s:%s as pointed to by snapshot: %s" % (symbol, version['version'],
                                                                                    snap_name))
                return
        self._versions.delete_one({'_id': version['_id']})
        # TODO: for FW pointers, if the above statement fails, they we have no way to delete the orphaned segments.
        #       This would be possible only via FSCK, or by moving the above statement at the end of this method,
        #       but with the risk of failing to delelte the version catastrophically, and ending up with a corrupted v.
        if do_cleanup:
            cleanup(self._arctic_lib, symbol, [version['_id']], self._versions,
                    shas_to_delete=tuple(bson.binary.Binary(s) for s in version.get(FW_POINTERS_REFS_KEY, [])),
                    pointers_cfgs=(get_fwptr_config(version), ))

    @mongo_retry
    def delete(self, symbol):
        """
        Delete all versions of the item from the current library which aren't
        currently part of some snapshot.

        Parameters
        ----------
        symbol : `str`
            symbol name to delete
        """
        logger.info("Deleting data item: %r from %r" % (symbol, self._arctic_lib.get_name()))
        # None is the magic sentinel value that indicates an item has been deleted.
        sentinel = self.write(symbol, None, prune_previous_version=False, metadata={'deleted': True})
        self._prune_previous_versions(symbol, 0)

        # If there aren't any other versions, then we don't need the sentinel empty value
        # so delete the sentinel version altogether
        snapped_version = self._versions.find_one({'symbol': symbol,
                                                   'metadata.deleted': {'$ne': True}})
        if not snapped_version:
            self._delete_version(symbol, sentinel.version)
        assert not self.has_symbol(symbol)

    def _write_audit(self, user, message, changed_version):
        """
        Creates an audit entry, which is much like a snapshot in that
        it references versions and provides some history of the changes made.
        """
        audit = {'_id': bson.ObjectId(),
                 'user': user,
                 'message': message,
                 'symbol': changed_version.symbol
                 }
        orig_version = changed_version.orig_version.version
        new_version = changed_version.new_version.version
        audit['orig_v'] = orig_version
        audit['new_v'] = new_version
        # Update the versions to contain the audit
        mongo_retry(self._versions.update_many)({'symbol': changed_version.symbol,
                                                 'version': {'$in': [orig_version, new_version]}
                                                 },
                                                {'$addToSet': {'parent': audit['_id']}})
        # Create the audit entry
        mongo_retry(self._audit.insert_one)(audit)

    @mongo_retry
    def snapshot(self, snap_name, metadata=None, skip_symbols=None, versions=None):
        """
        Snapshot versions of symbols in the library.  Can be used like:

        Parameters
        ----------
        snap_name : `str`
            name of the snapshot
        metadata : `dict`
            an optional dictionary of metadata to persist along with the symbol.
        skip_symbols : `collections.Iterable`
            optional symbols to be excluded from the snapshot
        versions: `dict`
            an optional dictionary of versions of the symbols to be snapshot
        """
        # Ensure the user doesn't insert duplicates
        snapshot = self._snapshots.find_one({'name': snap_name})
        if snapshot:
            raise DuplicateSnapshotException("Snapshot '%s' already exists." % snap_name)

        # Create a snapshot version document
        snapshot = {'_id': bson.ObjectId()}
        snapshot['name'] = snap_name
        snapshot['metadata'] = metadata

        skip_symbols = set() if skip_symbols is None else set(skip_symbols)

        if versions is None:
            versions = {sym: None for sym in set(self.list_symbols()) - skip_symbols}

        # Loop over, and snapshot all versions except those we've been asked to skip
        for sym in versions:
            try:
                sym = self._read_metadata(sym, read_preference=ReadPreference.PRIMARY, as_of=versions[sym])
                # Update the parents field of the version document
                mongo_retry(self._versions.update_one)({'_id': sym['_id']},
                                                       {'$addToSet': {'parent': snapshot['_id']}})
            except NoDataFoundException:
                # Version has been deleted, not included in the snapshot
                pass

        mongo_retry(self._snapshots.insert_one)(snapshot)

    @mongo_retry
    def delete_snapshot(self, snap_name):
        """
        Delete a named snapshot

        Parameters
        ----------
        symbol : `str`
            The snapshot name to delete
        """
        snapshot = self._snapshots.find_one({'name': snap_name})
        if not snapshot:
            raise NoDataFoundException("Snapshot %s not found!" % snap_name)

        # Remove the snapshot Id as a parent of versions
        self._versions.update_many({'parent': snapshot['_id']},
                                   {'$pull': {'parent': snapshot['_id']}})

        self._snapshots.delete_one({'name': snap_name})

    @mongo_retry
    def list_snapshots(self):
        """
        List the snapshots in the library

        Returns
        -------
        string list of snapshot names
        """
        return dict((i['name'], i['metadata']) for i in self._snapshots.find())

    @mongo_retry
    def stats(self):
        """
        Return storage statistics about the library

        Returns
        -------
        dictionary of storage stats
        """

        res = {}
        db = self._collection.database
        conn = db.connection
        res['sharding'] = {}
        try:
            sharding = conn.config.databases.find_one({'_id': db.name})
            if sharding:
                res['sharding'].update(sharding)
            res['sharding']['collections'] = list(conn.config.collections.find({'_id': {'$regex': '^' + db.name + r"\..*"}}))
        except OperationFailure:
            # Access denied
            pass
        res['dbstats'] = db.command('dbstats')
        res['chunks'] = db.command('collstats', self._collection.name)
        res['versions'] = db.command('collstats', self._versions.name)
        res['snapshots'] = db.command('collstats', self._snapshots.name)
        res['totals'] = {'count': res['chunks']['count'],
                         'size': res['chunks']['size'] + res['versions']['size'] + res['snapshots']['size'],
                         }
        return res

    def _fsck(self, dry_run):
        """
        Run a consistency check on this VersionStore library.
        """
        # Cleanup Orphaned Chunks
        self._cleanup_orphaned_chunks(dry_run)
        # Cleanup unreachable SHAs (forward pointers)
        self._cleanup_unreachable_shas(dry_run)
        # Cleanup Orphaned Snapshots
        self._cleanup_orphaned_versions(dry_run)

    def _cleanup_unreachable_shas(self, dry_run):
        lib = self
        chunks_coll = lib._collection
        versions_coll = chunks_coll.versions

        for symbol in chunks_coll.distinct('symbol'):
            logger.debug('Checking %s (forward pointers)' % symbol)

            all_symbol_pointers_cfgs = _get_symbol_pointer_cfgs(symbol, versions_coll)

            if FwPointersCfg.DISABLED not in all_symbol_pointers_cfgs:
                # Obtain the SHAs which are no longer pointed to by any version
                symbol_alive_shas = get_symbol_alive_shas(symbol, versions_coll)
                all_symbol_shas = set(chunks_coll.distinct('sha', {'symbol': symbol}))
                unreachable_shas = all_symbol_shas - symbol_alive_shas

                logger.info("Cleaning up {} SHAs for symbol {}".format(len(unreachable_shas), symbol))
                if not dry_run:
                    # Be liberal with the generation time.
                    id_time_constraint = {'$lt': bson.ObjectId.from_datetime(dt.now() - timedelta(days=1))}
                    # Do delete the data segments
                    chunks_coll.delete_many({
                        '_id': id_time_constraint,  # can't rely on the parent field only for fw-pointers
                        'symbol': symbol,
                        'parent': id_time_constraint,
                        'sha': {'$in': list(unreachable_shas)}})

    def _cleanup_orphaned_chunks(self, dry_run):
        """
        Fixes any chunks who have parent pointers to missing versions.
        Removes the broken parent pointer and, if there are no other parent pointers for the chunk,
        removes the chunk.
        """
        lib = self
        chunks_coll = lib._collection
        versions_coll = chunks_coll.versions

        logger.info("ORPHANED CHUNK CHECK: %s" % self._arctic_lib.get_name())
        for symbol in chunks_coll.distinct('symbol'):
            logger.debug('Checking %s' % symbol)
            # Be liberal with the generation time.
            gen_time = dt.now() - timedelta(days=1)
            parent_id_constraint = {'$lt': bson.ObjectId.from_datetime(gen_time)}

            # For each symbol, grab all 'real' versions
            versions = set(versions_coll.find({'symbol': symbol,
                                               '_id': parent_id_constraint}).distinct('_id'))
            # Using aggregate so we can unwind, and pull out 'parent', where 'parent' is older than a day.
            parents = chunks_coll.aggregate([{'$match': {'symbol': symbol}},
                                             {'$project': {'parent': True}},
                                             {'$unwind': '$parent'},
                                             {'$match': {'parent': parent_id_constraint}},
                                             {'$group': {'_id': '$parent'}},
                                             ])
            parent_ids = set([x['_id'] for x in parents])

            leaked_versions = sorted(parent_ids - versions)
            if len(leaked_versions):
                logger.info("%s leaked %d versions" % (symbol, len(leaked_versions)))
            for x in leaked_versions:
                chunk_count = mongo_count(chunks_coll, filter={'symbol': symbol, 'parent': x})
                logger.info("%s: Missing Version %s (%s) ; %s chunks ref'd" % (symbol,
                                                                               x.generation_time,
                                                                               x,
                                                                               chunk_count
                                                                               ))
                if versions_coll.find_one({'symbol': symbol, '_id': x}) is not None:
                    raise Exception("Error: version (%s) is found for (%s), but shouldn't be!" %
                                    (x, symbol))
            # Now cleanup the leaked versions
            if not dry_run:
                # This is now able to handle safely symbols which have both forward and legacy/parent pointers
                cleanup(lib._arctic_lib, symbol, leaked_versions, versions_coll)

    def _cleanup_orphaned_versions(self, dry_run):
        """
        Fixes any versions who have parent pointers to missing snapshots.
        Note, doesn't delete the versions, just removes the parent pointer if it no longer
        exists in snapshots.
        """
        lib = self
        versions_coll = lib._collection.versions
        snapshots_coll = lib._collection.snapshots

        logger.info("ORPHANED SNAPSHOT CHECK: %s" % self._arctic_lib.get_name())

        # Be liberal with the generation time.
        gen_time = dt.now() - timedelta(days=1)
        parent_id_constraint = {'$lt': bson.ObjectId.from_datetime(gen_time)}

        # For each symbol, grab all 'real' snapshots and audit entries
        snapshots = set(snapshots_coll.distinct('_id'))
        snapshots |= set(lib._audit.distinct('_id'))
        # Using aggregate so we can unwind, and pull out 'parent', where 'parent' is older than a day.
        parents = versions_coll.aggregate([{'$project': {'parent': True}},
                                           {'$unwind': '$parent'},
                                           {'$match': {'parent': parent_id_constraint}},
                                           {'$group': {'_id': '$parent'}},
                                           ])
        parent_ids = set([x['_id'] for x in parents])

        leaked_snaps = sorted(parent_ids - snapshots)
        if len(leaked_snaps):
            logger.info("leaked %d snapshots" % (len(leaked_snaps)))
        for x in leaked_snaps:
            ver_count = mongo_count(versions_coll, filter={'parent': x})
            logger.info("Missing Snapshot %s (%s) ; %s versions ref'd" % (x.generation_time,
                                                                          x,
                                                                          ver_count
                                                                          ))
            if snapshots_coll.find_one({'_id': x}) is not None:
                raise Exception("Error: snapshot (%s) is found, but shouldn't be!" %
                                (x))
            # Now cleanup the leaked snapshots
            if not dry_run:
                versions_coll.update_many({'parent': x},
                                          {'$pull': {'parent': x}})



import pymongo
import pytest
from pymongo.errors import OperationFailure

import arctic
from arctic.store import _version_store_utils as vsu
from tests.integration.chunkstore.test_utils import create_test_data

symbol = 'TS1'


def n_append(library, library_name, total_appends, rows_per_append, bulk_data_ts, start_idx, do_snapshots=True, do_prune=True):
    open_last_row = 0
    for i in range(total_appends):
        first_row = start_idx + i * rows_per_append
        open_last_row = start_idx + (i + 1) * rows_per_append
        snap = 'snap_{}'.format(first_row)
        library.append(symbol, bulk_data_ts[first_row:open_last_row],
                       metadata={'snap': snap},
                       prune_previous_version=do_prune)
        if do_snapshots:
            library.snapshot(snap)

    return open_last_row


def _corrupt_with_append_only(library, library_name):
    def do_fail(version):
        raise Exception('test')

    large_ts = create_test_data(size=2000, cols=100,
                                index=True, multiindex=False,
                                random_data=True, random_ids=True)
    library.write(symbol, large_ts[0:1000])  # v1
    library.snapshot('snap_write_a')
    library.append(symbol, large_ts[1000:1010])  # v2
    library.snapshot('snap_write_b')

    # Here we simulate a scenario where an append succeeds to insert the data segments,
    # but fails to insert the version document (i.e. Mongo error occurred)
    orig_insert_version = library._insert_version
    library._insert_version = do_fail
    try:
        library.append(symbol, large_ts[1010:1020])  # v3
    except:
        pass
    library._insert_version = orig_insert_version

    library.write_metadata(symbol, {'hello': 'there'})  # , prune_previous_version=False)

    # Appending subsequently overlapping and non-SHA-matching data cause data corruption
    library.append(symbol, large_ts[1018:1030])  # , prune_previous_version=False)

    last_v = library._versions.find_one(sort=[('version', pymongo.DESCENDING)])
    vsu.analyze_symbol(library, symbol, 0, last_v['version'] + 1)

    # Verify no versions have been corrupted
    for v in library._versions.find(sort=[('version', pymongo.DESCENDING)]):
        library.read(symbol, as_of=v['version'])


def test_no_corruption_restore_append_overlapping(library, library_name):
    large_ts = create_test_data(size=3000, cols=100,
                                index=True, multiindex=False,
                                random_data=True, random_ids=True)
    rows_per_append = 100

    n_append(library, library_name, 18, rows_per_append, large_ts, 0)

    # Corrupts all versions between the version that row "restore_from_row" was written,
    restore_from_row = rows_per_append * 10
    library.restore_version(symbol, 'snap_{}'.format(restore_from_row))
    library.append(symbol, large_ts[restore_from_row:restore_from_row + 50])

    last_v = library._versions.find_one(sort=[('version', pymongo.DESCENDING)])
    vsu.analyze_symbol(library, symbol, 0, last_v['version'] + 1)

    # Verify no versions have been corrupted
    for v in library._versions.find(sort=[('version', pymongo.DESCENDING)]):
        library.read(symbol, as_of=v['version'])


def test_no_corruption_restore_writemeta_append(library, library_name):
    large_ts = create_test_data(size=2000, cols=100,
                                index=True, multiindex=False,
                                random_data=True, random_ids=True)
    rows_per_append = 100

    last_row = n_append(library, library_name, 9, rows_per_append, large_ts, 0)

    library.write_metadata(symbol, metadata={'abc': 'xyz'})

    n_append(library, library_name, 9, rows_per_append, large_ts, last_row)

    library.write_metadata(symbol, metadata={'abc2': 'xyz2'})

    # Corrupts all versions between the version that row "restore_from_row" was written,
    restore_from_row = rows_per_append * 10
    library.restore_version(symbol, 'snap_{}'.format(restore_from_row))

    library.write_metadata(symbol, metadata={'abc3': 'xyz3'})

    library.append(symbol, large_ts[restore_from_row:restore_from_row + 50])

    library.write_metadata(symbol, metadata={'abc4': 'xyz4'})

    last_v = library._versions.find_one(sort=[('version', pymongo.DESCENDING)])
    vsu.analyze_symbol(library, symbol, 0, last_v['version'] + 1)

    # Verify no versions have been corrupted
    for v in library._versions.find(sort=[('version', pymongo.DESCENDING)]):
        library.read(symbol, as_of=v['version'])


def test_no_corruption_restore_append_non_overlapping_tstamps(library, library_name):
    large_ts = create_test_data(size=2000, cols=100,
                                index=True, multiindex=False,
                                random_data=True, random_ids=True)

    # Append with 50 small uncompressed segments (no new base yet)
    last_row_b = n_append(library, library_name, 50, 25, large_ts, 0, False, True)

    library.snapshot('snap_A')

    # Append with 20 more small segments, causes once copy-rewrite with new base, and then some small appended segments
    n_append(library, library_name, 15, 25, large_ts, last_row_b, True, True)

    library.restore_version(symbol, as_of='snap_A')

    last_row = n_append(library, library_name, 1, 40, large_ts, last_row_b, False, True)
    library.snapshot('snap_B')

    # Corrupts all versions
    last_row = n_append(library, library_name, 1, 10, large_ts, last_row, False, True)
    last_row = n_append(library, library_name, 8, 20, large_ts, last_row, False, True)
    library.snapshot('snap_C')

    last_v = library._versions.find_one(sort=[('version', pymongo.DESCENDING)])
    vsu.analyze_symbol(library, symbol, 0, last_v['version'] + 1)

    # Verify no versions have been corrupted
    for v in library._versions.find(sort=[('version', pymongo.DESCENDING)]):
        library.read(symbol, as_of=v['version'])


def test_restore_append_overlapping_corrupts_old(library, library_name):
    large_ts = create_test_data(size=2000, cols=100,
                                index=True, multiindex=False,
                                random_data=True, random_ids=True)
    library.write(symbol, large_ts[0:1000])
    library.snapshot('snap_write_a')

    library.append(symbol, large_ts[1000:1010])

    library.restore_version(symbol, as_of='snap_write_a', prune_previous_version=True)
    library.append(symbol, large_ts[1000:1009])

    last_v = library._versions.find_one(sort=[('version', pymongo.DESCENDING)])
    vsu.analyze_symbol(library, symbol, 0, last_v['version'] + 1)

    # Verify no versions have been corrupted
    for v in library._versions.find(sort=[('version', pymongo.DESCENDING)]):
        library.read(symbol, as_of=v['version'])


def test_restore_append_overlapping_corrupts_last(library, library_name):
    large_ts = create_test_data(size=2000, cols=100,
                                index=True, multiindex=False,
                                random_data=True, random_ids=True)
    library.write(symbol, large_ts[0:1000])
    library.snapshot('snap_write_a')

    library.append(symbol, large_ts[1000:1010])

    library.restore_version(symbol, as_of='snap_write_a', prune_previous_version=True)
    library.append(symbol, large_ts[1000:1012])

    last_v = library._versions.find_one(sort=[('version', pymongo.DESCENDING)])
    vsu.analyze_symbol(library, symbol, 0, last_v['version'] + 1)

    # Verify no versions have been corrupted
    for v in library._versions.find(sort=[('version', pymongo.DESCENDING)]):
        library.read(symbol, as_of=v['version'])


# This is not necessary to fix, but the Exception thrown is quite confusing.
@pytest.mark.skip(reason="Not critical as upsert=False is rarely used. A more specific handling/exception is required here.")
def test_append_fail_after_delete_noupsert(library, library_name):
    large_ts = create_test_data(size=2000, cols=100,
                                index=True, multiindex=False,
                                random_data=True, random_ids=True)
    library.write(symbol, large_ts[0:1000])  #v1
    library.snapshot('snap_a')
    library.append(symbol, large_ts[1000:1010])  #v2
    library.snapshot('snap_b')
    library.append(symbol, large_ts[1010:1020])  #v3
    library.snapshot('snap_c')

    library.append(symbol, large_ts[1030:1040])  #v4

    library.delete(symbol) #v5

    library.append(symbol, large_ts[1040:1050], upsert=False)  # v6

    last_v = library._versions.find_one(sort=[('version', pymongo.DESCENDING)])
    vsu.analyze_symbol(library, symbol, 0, last_v['version'] + 1)

    # Verify no versions have been corrupted
    for v in library._versions.find(sort=[('version', pymongo.DESCENDING)]):
        library.read(symbol, as_of=v['version'])


def test_append_without_corrupt_check(library, library_name):
    orig_check = arctic.store._ndarray_store.CHECK_CORRUPTION_ON_APPEND
    arctic.store._ndarray_store.set_corruption_check_on_append(False)
    try:
        with pytest.raises(OperationFailure):
            _corrupt_with_append_only(library, library_name)
    finally:
        arctic.store._ndarray_store.set_corruption_check_on_append(orig_check)


def test_append_with_corrupt_check(library, library_name):
    orig_check = arctic.store._ndarray_store.CHECK_CORRUPTION_ON_APPEND
    arctic.store._ndarray_store.set_corruption_check_on_append(True)
    try:
        _corrupt_with_append_only(library, library_name)
    finally:
        arctic.store._ndarray_store.set_corruption_check_on_append(orig_check)


def test_fast_check_corruption(library, library_name):
    ts = create_test_data(size=100, cols=100,
                          index=True, multiindex=False,
                          random_data=True, random_ids=True)
    library.write(symbol, ts[0:10])  # v1

    assert not vsu.fast_is_corrupted(library, symbol, input_v=1)

    library.append(symbol, ts[10:20], prune_previous_version=False)  # v2
    assert not vsu.fast_is_corrupted(library, symbol, input_v=2)

    library.append(symbol, ts[20:30], prune_previous_version=False)  # v3
    assert not vsu.fast_is_corrupted(library, symbol, input_v=3)

    # Now the dangerous part
    last_segment = library._collection.find_one({}, sort=[('_id', pymongo.DESCENDING)])
    library._collection.delete_one({'_id': last_segment['_id']})

    assert vsu.fast_is_corrupted(library, symbol, input_v=3)


def test_fast_is_safe_to_append(library, library_name):
    from bson.binary import Binary
    import hashlib
    def modify_segment(segment, item):
        segment['segment'] -= 2
        sha = hashlib.sha1()
        sha.update(item.encode('ascii'))
        segment['sha'] = Binary(sha.digest())
        segment.pop('_id')

    ts = create_test_data(size=100, cols=100,
                          index=True, multiindex=False,
                          random_data=True, random_ids=True)
    library.write(symbol, ts[0:10])  # v1
    assert vsu.is_safe_to_append(library, symbol, input_v=1)

    library.append(symbol, ts[10:20], prune_previous_version=False)  # v2
    assert vsu.is_safe_to_append(library, symbol, input_v=2)

    library.append(symbol, ts[20:30], prune_previous_version=False)  # v3
    assert vsu.is_safe_to_append(library, symbol, input_v=3)

    # Corrupt the data be removing segment
    last_segment = library._collection.find_one({}, sort=[('_id', pymongo.DESCENDING)])
    library._collection.delete_one({'_id': last_segment['_id']})
    assert not vsu.is_safe_to_append(library, symbol, input_v=3)
    with pytest.raises(OperationFailure):
        library.read(symbol)

    # Fix the library by adding back the deleted segment
    library._collection.insert_one(last_segment)
    assert vsu.is_safe_to_append(library, symbol, input_v=3)

    # Corrupt the data be adding an unnecessary segment
    modify_segment(last_segment, 'abcd')
    library._collection.insert_one(last_segment)
    assert not vsu.is_safe_to_append(library, symbol, input_v=3)
    with pytest.raises(OperationFailure):
        library.read(symbol)



import pytest

import arctic.serialization.numpy_records as anr
from tests.unit.serialization.serialization_test_data import _mixed_test_data as input_test_data

df_serializer = anr.DataFrameSerializer()


@pytest.mark.parametrize("input_df", input_test_data().keys())
def test_dataframe_confirm_fast_check_compatibility(input_df):
    orig_config = anr.FAST_CHECK_DF_SERIALIZABLE
    try:
        input_df = input_test_data()[input_df][0]
        anr.set_fast_check_df_serializable(True)
        with_fast_check = df_serializer.can_convert_to_records_without_objects(input_df, 'symA')
        anr.set_fast_check_df_serializable(False)
        without_fast_check = df_serializer.can_convert_to_records_without_objects(input_df, 'symA')
        assert with_fast_check == without_fast_check
    finally:
        anr.FAST_CHECK_DF_SERIALIZABLE = orig_config



import abc
import hashlib
import logging
from threading import RLock

import numpy as np
import pandas as pd
from bson import Binary

from arctic._config import ARCTIC_AUTO_EXPAND_CHUNK_SIZE
from arctic.serialization.numpy_records import PandasSerializer
from .._compression import compress
from .._config import MAX_DOCUMENT_SIZE
from .._util import NP_OBJECT_DTYPE
from ..exceptions import ArcticSerializationException

ABC = abc.ABCMeta('ABC', (object,), {})

log = logging.getLogger(__name__)


def incremental_checksum(item, curr_sha=None):
    curr_sha = hashlib.sha1() if curr_sha is None else curr_sha
    curr_sha.update(item)
    return curr_sha


class LazyIncrementalSerializer(ABC):
    def __init__(self, serializer, input_data, chunk_size):
        if chunk_size < 1:
            raise ArcticSerializationException("LazyIncrementalSerializer can't be initialized "
                                               "with chunk_size < 1 ({})".format(chunk_size))
        if not serializer:
            raise ArcticSerializationException("LazyIncrementalSerializer can't be initialized "
                                               "with a None serializer object")
        self.input_data = input_data
        self.chunk_size = chunk_size
        self._serializer = serializer
        self._initialized = False
        self._checksum = None

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractproperty
    def generator(self):
        pass

    @abc.abstractproperty
    def generator_bytes(self):
        pass

    @abc.abstractproperty
    def serialize(self):
        pass


class IncrementalPandasToRecArraySerializer(LazyIncrementalSerializer):
    def __init__(self, serializer, input_data, chunk_size, string_max_len=None):
        super(IncrementalPandasToRecArraySerializer, self).__init__(serializer, input_data, chunk_size)
        if not isinstance(serializer, PandasSerializer):
            raise ArcticSerializationException("IncrementalPandasToRecArraySerializer requires a serializer of "
                                               "type PandasSerializer.")
        if not isinstance(input_data, (pd.DataFrame, pd.Series)):
            raise ArcticSerializationException("IncrementalPandasToRecArraySerializer requires a pandas DataFrame or "
                                               "Series as data source input.")
        if string_max_len and string_max_len < 1:
            raise ArcticSerializationException("IncrementalPandasToRecArraySerializer can't be initialized "
                                               "with string_max_len < 1 ({})".format(string_max_len))
        self.string_max_len = string_max_len
        # The state which needs to be lazily initialized
        self._dtype = None
        self._shape = None
        self._rows_per_chunk = 0
        self._total_chunks = 0
        self._has_string_object = False
        self._lock = RLock()

    def _dtype_convert_to_max_len_string(self, input_ndtype, fname):
        if input_ndtype.type not in (np.string_, np.unicode_):
            return input_ndtype, False
        type_sym = 'S' if input_ndtype.type == np.string_ else 'U'
        max_str_len = len(max(self.input_data[fname].astype(type_sym), key=len))
        str_field_dtype = np.dtype('{}{:d}'.format(type_sym, max_str_len)) if max_str_len > 0 else input_ndtype
        return str_field_dtype, True

    def _get_dtype(self):
        # Serializer is being called only if can_convert_to_records_without_objects() has passed,
        # which means that the resulting recarray does not contain objects but only numpy types, string, or unicode

        # Serialize the first row to obtain info about row size in bytes (cache first few rows only)
        # Also raise an Exception early, if data are not serializable
        first_chunk, serialized_dtypes = self._serializer.serialize(
            self.input_data[0:10] if len(self) > 0 else self.input_data,
            string_max_len=self.string_max_len)

        # This is the common case, where first row's dtype represents well the whole dataframe's dtype
        if serialized_dtypes is None or \
                len(self.input_data) == 0 or \
                NP_OBJECT_DTYPE not in self.input_data.dtypes.values:
            return first_chunk, serialized_dtypes, False

        # Reaching here means we have at least one column of type object
        # To correctly serialize incrementally, we need to know the final dtype (type and fixed length),
        # using length-conversion information from all values of the object columns

        dtype_arr = []
        has_string_object = False
        for field_name in serialized_dtypes.names:  # include all column names, along with the expanded multi-index
            field_dtype = serialized_dtypes[field_name]
            if field_name not in self.input_data or self.input_data.dtypes[field_name] is NP_OBJECT_DTYPE:
                # Note: .hasobject breaks for timezone-aware datetime64 pandas columns, so compare with dtype('O')
                # if column is an expanded multi index or doesn't contain objects, the serialized 1st row dtype is safe
                field_dtype, with_str_object = self._dtype_convert_to_max_len_string(field_dtype, field_name)
                has_string_object |= with_str_object
            dtype_arr.append((field_name, field_dtype))
        return first_chunk, np.dtype(dtype_arr), has_string_object

    def _lazy_init(self):
        if self._initialized:
            return

        with self._lock:
            if self._initialized:  # intentional double check here
                return
            # Get the dtype of the serialized array (takes into account object types, converted to fixed length strings)
            first_chunk, dtype, has_string_object = self._get_dtype()

            # Compute the number of rows which can fit in a chunk
            rows_per_chunk = 0
            if len(self) > 0 and self.chunk_size > 1:
                rows_per_chunk = IncrementalPandasToRecArraySerializer._calculate_rows_per_chunk(self.chunk_size, first_chunk)

            # Initialize object's state
            self._dtype = dtype
            shp = list(first_chunk.shape)
            shp[0] = len(self)
            self._shape = tuple(shp)
            self._has_string_object = has_string_object
            self._rows_per_chunk = rows_per_chunk
            self._total_chunks = int(np.ceil(float(len(self)) / self._rows_per_chunk)) if rows_per_chunk > 0 else 0
            self._initialized = True

    @staticmethod
    def _calculate_rows_per_chunk(max_chunk_size, chunk):
        sze = int(chunk.dtype.itemsize * np.prod(chunk.shape[1:]))
        sze = sze if sze < max_chunk_size else max_chunk_size
        rows_per_chunk = int(max_chunk_size / sze)
        if rows_per_chunk < 1 and ARCTIC_AUTO_EXPAND_CHUNK_SIZE:
            # If a row size is larger than chunk_size, use the maximum document size
            logging.warning('Chunk size of {} is too small to fit a row ({}). '
                            'Using maximum document size.'.format(max_chunk_size, MAX_DOCUMENT_SIZE))
            # For huge rows, fall-back to using a very large document size, less than max-allowed by MongoDB
            rows_per_chunk = int(MAX_DOCUMENT_SIZE / sze)
        if rows_per_chunk < 1:
            raise ArcticSerializationException("Serialization failed to split data into max sized chunks.")
        return rows_per_chunk

    def __len__(self):
        return len(self.input_data)

    @property
    def shape(self):
        self._lazy_init()
        return self._shape

    @property
    def dtype(self):
        self._lazy_init()
        return self._dtype

    @property
    def rows_per_chunk(self):
        self._lazy_init()
        return self._rows_per_chunk

    def checksum(self, from_idx, to_idx):
        if self._checksum is None:
            self._lazy_init()
            total_sha = None
            for chunk_bytes, dtype in self.generator_bytes(from_idx=from_idx, to_idx=to_idx):
                # TODO: what about compress_array here in batches?
                compressed_chunk = compress(chunk_bytes)
                total_sha = incremental_checksum(compressed_chunk, curr_sha=total_sha)
            self._checksum = Binary(total_sha.digest())
        return self._checksum

    def generator(self, from_idx=None, to_idx=None):
        return self._generator(from_idx=from_idx, to_idx=to_idx)

    def generator_bytes(self, from_idx=None, to_idx=None):
        return self._generator(from_idx=from_idx, to_idx=to_idx, get_bytes=True)

    def _generator(self, from_idx, to_idx, get_bytes=False):
        # Note that the range is: [from_idx, to_idx)
        self._lazy_init()

        my_length = len(self)

        # Take into account default arguments and negative indexing (from end offset)
        from_idx = 0 if from_idx is None else from_idx
        if from_idx < 0:
            from_idx = my_length + from_idx
        to_idx = my_length if to_idx is None else min(to_idx, my_length)
        if to_idx < 0:
            to_idx = my_length + to_idx

        # No data, finish iteration
        if my_length == 0 or from_idx >= my_length or from_idx >= to_idx:
            return

        # Perform serialization for each chunk
        while from_idx < to_idx:
            curr_stop = min(from_idx + self._rows_per_chunk, to_idx)

            chunk, _ = self._serializer.serialize(
                self.input_data[from_idx: curr_stop],
                string_max_len=self.string_max_len,
                forced_dtype=self.dtype if self._has_string_object else None)

            # Let the gc collect the intermediate serialized chunk as early as possible
            chunk = chunk.tobytes() if chunk is not None and get_bytes else chunk

            yield chunk, self.dtype, from_idx, curr_stop
            from_idx = curr_stop

    def serialize(self):
        return self._serializer.serialize(self.input_data, self.string_max_len)



import abc
import logging
import uuid
from threading import RLock, Event

from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, FIRST_EXCEPTION

from arctic._config import ARCTIC_ASYNC_NWORKERS
from arctic.exceptions import AsyncArcticException

ABC = abc.ABCMeta('ABC', (object,), {})


def _looping_task(shutdown_flag, fun, *args, **kwargs):
    while not shutdown_flag.is_set():
        try:
            fun(*args, **kwargs)
        except Exception as e:
            logging.exception("Task failed {}".format(fun))
            raise e


def _exec_task(fun, *args, **kwargs):
    try:
        fun(*args, **kwargs)
    except Exception as e:
        logging.exception("Task failed {}".format(fun))
        raise e


class LazySingletonTasksCoordinator(ABC):
    """
    A Thread-Safe singleton lazily initialized thread pool class (encapsulating concurrent.futures.ThreadPoolExecutor)
    """
    _instance = None
    _SINGLETON_LOCK = RLock()
    _POOL_LOCK = RLock()

    @classmethod
    def is_initialized(cls):
        with cls._POOL_LOCK:
            is_init = cls._instance is not None and cls._instance._pool is not None
        return is_init

    @classmethod
    def get_instance(cls, pool_size=None):
        if cls._instance is not None:
            return cls._instance

        # Lazy init
        with cls._SINGLETON_LOCK:
            if cls._instance is None:
                cls._instance = cls(ARCTIC_ASYNC_NWORKERS if pool_size is None else pool_size)
        return cls._instance

    @property
    def _workers_pool(self):
        if self._pool is not None:
            return self._pool

        # lazy init the workers pool
        got_initialized = False
        with type(self)._POOL_LOCK:
            if self._pool is None:
                self._pool = ThreadPoolExecutor(max_workers=self._pool_size,
                                                thread_name_prefix='AsyncArcticWorker')
                got_initialized = True

        # Call hooks outside the lock, to minimize time-under-lock
        if got_initialized:
            for hook in self._pool_update_hooks:
                hook(self._pool_size)

        return self._pool

    def __init__(self, pool_size):
        # Only allow creation via get_instance
        if not type(self)._SINGLETON_LOCK._is_owned():
            raise AsyncArcticException("{} is a singleton, can't create a new instance".format(type(self)))

        pool_size = int(pool_size)
        if pool_size < 1:
            raise ValueError("{} can't be instantiated with a pool_size of {}".format(type(self), pool_size))

        # Enforce the singleton pattern
        with type(self)._SINGLETON_LOCK:
            if type(self)._instance is not None:
                raise AsyncArcticException("LazySingletonTasksCoordinator is a singleton, can't create a new instance")
            self._lock = RLock()
            self._pool = None
            self._pool_size = int(pool_size)
            self._pool_update_hooks = []
            self.alive_tasks = {}
            self.is_shutdown = False

    def reset(self, pool_size=None, timeout=None):
        pool_size = ARCTIC_ASYNC_NWORKERS if pool_size is None else int(pool_size)
        with type(self)._POOL_LOCK:
            self.shutdown(timeout=timeout)
            pool_size = max(pool_size, 1)
            self._pool = None
            self._pool_size = pool_size
            self.is_shutdown = False
            # pool will be lazily initialized with pool_size on next request submission

    def stop_all_running_tasks(self):
        with type(self)._POOL_LOCK:
            for fut, ev in (v for v in self.alive_tasks.values() if not v[0].done()):
                if ev:
                    ev.set()
                fut.cancel()

    @staticmethod
    def wait_tasks(futures, timeout=None, return_when=ALL_COMPLETED, raise_exceptions=True):
        running_futures = [fut for fut in futures if not fut.done()]
        done, _ = wait(running_futures, timeout=timeout, return_when=return_when)
        if raise_exceptions:
            [f.result() for f in done if not f.cancelled() and f.exception() is not None]  # raises the exception

    @staticmethod
    def wait_tasks_or_abort(futures, timeout=60, kill_switch_ev=None):
        try:
            LazySingletonTasksCoordinator.wait_tasks(futures, return_when=FIRST_EXCEPTION, raise_exceptions=True)
        except Exception as e:
            if kill_switch_ev is not None:
                # Used when we want to keep both raise the exception and wait for all tasks to finish
                kill_switch_ev.set()
                LazySingletonTasksCoordinator.wait_tasks(futures, return_when=ALL_COMPLETED,
                                                         raise_exceptions=False, timeout=timeout)
            raise e

    def register_update_hook(self, fun):
        with type(self)._POOL_LOCK:
            self._pool_update_hooks.append(fun)

    def submit_task(self, is_looping, fun, *args, **kwargs):
        new_id = uuid.uuid4()
        shutdown_flag = Event() if is_looping else None
        with type(self)._POOL_LOCK:
            if self.is_shutdown:
                raise AsyncArcticException("The worker pool has been shutdown and can no longer accept new requests.")

            if is_looping:
                new_future = self._workers_pool.submit(_looping_task, shutdown_flag, fun, *args, **kwargs)
            else:
                new_future = self._workers_pool.submit(_exec_task, fun, *args, **kwargs)
            self.alive_tasks = {k: v for k, v in self.alive_tasks.items() if not v[0].done()}
            self.alive_tasks[new_id] = (new_future, shutdown_flag)
        return new_id, new_future

    def total_alive_tasks(self):
        with type(self)._POOL_LOCK:
            self.alive_tasks = {k: v for k, v in self.alive_tasks.items() if not v[0].done()}
            total = len(self.alive_tasks)
        return total

    def shutdown(self, timeout=None):
        if self.is_shutdown:
            return
        with type(self)._POOL_LOCK:
            self.is_shutdown = True
        if timeout is not None:
            self.await_termination(timeout=timeout)
        self._workers_pool.shutdown(wait=timeout is not None)

    def await_termination(self, timeout=None):
        with type(self)._POOL_LOCK:
            if not self.is_shutdown:
                raise AsyncArcticException("The workers pool has not been shutdown, please call shutdown() first.")
        LazySingletonTasksCoordinator.wait_tasks(
            [v[0] for v in self.alive_tasks.values()],
            timeout=timeout, return_when=ALL_COMPLETED, raise_exceptions=False)
        with type(self)._POOL_LOCK:
            self.alive_tasks = {}

    @property
    def actual_pool_size(self):
        return self._workers_pool._max_workers

    @abc.abstractmethod
    def __reduce__(self):
        pass



import logging
from multiprocessing.pool import ThreadPool

try:
    from lz4.block import compress as lz4_compress, decompress as lz4_decompress
    lz4_compressHC = lambda _str: lz4_compress(_str, mode='high_compression')
except ImportError as e:
    from lz4 import compress as lz4_compress, compressHC as lz4_compressHC, decompress as lz4_decompress

# ENABLE_PARALLEL mutated in global_scope. Do not remove.
from ._config import ENABLE_PARALLEL, LZ4_HIGH_COMPRESSION, LZ4_WORKERS, LZ4_N_PARALLEL, LZ4_MINSZ_PARALLEL, \
    BENCHMARK_MODE  # noqa # pylint: disable=unused-import

logger = logging.getLogger(__name__)


_compress_thread_pool = None


def enable_parallel_lz4(mode):
    """
    Set the global multithread compression mode

    Parameters
    ----------
        mode: `bool`
            True: Use parallel compression. False: Use sequential compression
    """
    global ENABLE_PARALLEL
    ENABLE_PARALLEL = bool(mode)
    logger.info("Setting parallelisation mode to {}".format("multi-threaded" if mode else "single-threaded"))


def set_compression_pool_size(pool_size):
    """
    Set the size of the compression workers thread pool.
    If the pool is already created, it waits until all jobs are finished, and then proceeds with setting the new size.

    Parameters
    ----------
        pool_size : `int`
            The size of the pool (must be a positive integer)

    Returns
    -------
    `None`
    """
    pool_size = int(pool_size)
    if pool_size < 1:
        raise ValueError("The compression thread pool size cannot be of size {}".format(pool_size))

    global _compress_thread_pool
    if _compress_thread_pool is not None:
        _compress_thread_pool.close()
        _compress_thread_pool.join()
    _compress_thread_pool = ThreadPool(pool_size)


def compress_array(str_list, withHC=LZ4_HIGH_COMPRESSION):
    """
    Compress an array of strings

    Parameters
    ----------
        str_list: `list[str]`
            The input list of strings which need to be compressed.
        withHC: `bool`
            This flag controls whether lz4HC will be used.

    Returns
    -------
    `list[str`
    The list of the compressed strings.
    """
    global _compress_thread_pool

    if not str_list:
        return str_list

    do_compress = lz4_compressHC if withHC else lz4_compress

    def can_parallelize_strlist(strlist):
        return len(strlist) > LZ4_N_PARALLEL and len(strlist[0]) > LZ4_MINSZ_PARALLEL

    use_parallel = (ENABLE_PARALLEL and withHC) or can_parallelize_strlist(str_list)

    if BENCHMARK_MODE or use_parallel:
        if _compress_thread_pool is None:
            _compress_thread_pool = ThreadPool(LZ4_WORKERS)
        return _compress_thread_pool.map(do_compress, str_list)

    return [do_compress(s) for s in str_list]


def compress(_str):
    """
    Compress a string

    By default LZ4 mode is standard in interactive mode,
    and high compresion in applications/scripts
    """
    return lz4_compress(_str)


def compressHC(_str):
    """
    HC compression
    """
    return lz4_compressHC(_str)


def compressHC_array(str_list):
    """
    HC compression
    """
    return compress_array(str_list, withHC=True)


def decompress(_str):
    """
    Decompress a string
    """
    return lz4_decompress(_str)


def decompress_array(str_list):
    """
    Decompress a list of strings
    """
    global _compress_thread_pool

    if not str_list:
        return str_list

    if not ENABLE_PARALLEL or len(str_list) <= LZ4_N_PARALLEL:
        return [lz4_decompress(chunk) for chunk in str_list]

    if _compress_thread_pool is None:
        _compress_thread_pool = ThreadPool(LZ4_WORKERS)
    return _compress_thread_pool.map(lz4_decompress, str_list)



from mock import Mock

import arctic.fixtures.arctic as fix
from arctic.arctic import Arctic


def test_overlay_library_name(overlay_library_name):
    assert(overlay_library_name == 'test.OVERLAY')


def test_overlay_library():
    a = Mock(Arctic, autospec=True)
    fix._overlay_library(a, 'test')
    a.initialize_library.assert_called_with("test_RAW", "VersionStore", segment='year')


def test_tickstore_lib():
    a = Mock(Arctic, autospec=True)
    fix._tickstore_lib(a, "test")
    a.initialize_library.assert_called_with('test', 'TickStoreV3')
    a.get_library.assert_called_with('test')



from mock import MagicMock, ANY, patch

from arctic._util import are_equals, enable_sharding, mongo_count
from arctic.arctic import Arctic
import arctic._util


def test_are_equals_not_df():
    assert(are_equals(1.0, 2.0) is False)
    assert(are_equals([1, 2, 3], [1, 2, 3]))
    assert(are_equals("Hello", "World") is False)


def test_enable_sharding_hashed():
    m = MagicMock(Arctic, autospec=True)
    enable_sharding(m, "test", hashed=True)
    m._conn.admin.command.assert_called_with('shardCollection', ANY, key={'symbol': 'hashed'})


def test_mongo_count_old_pymongo(monkeypatch):
    monkeypatch.setattr(arctic._util, '_use_new_count_api', None)
    with patch('pymongo.version', '3.6.0'):
        coll = MagicMock()
        mongo_count(coll, filter="_id:1")
        mongo_count(coll, filter={})
        mongo_count(coll)
        assert coll.estimated_document_count.call_count == 0
        assert coll.count_documents.call_count == 0
        assert coll.count.call_count == 3


def test_mongo_count_new_pymongo(monkeypatch):
    monkeypatch.setattr(arctic._util, '_use_new_count_api', None)
    with patch('pymongo.version', '3.11.0'):
        coll2 = MagicMock()
        mongo_count(coll2, filter="_id:1")
        mongo_count(coll2, filter={})
        mongo_count(coll2)
        assert coll2.estimated_document_count.call_count == 2
        assert coll2.count_documents.call_count == 1
        assert coll2.count.call_count == 0



try:
    from unittest.mock import Mock, patch, MagicMock, create_autospec, sentinel, call
except:    
    from mock import Mock, patch, MagicMock, create_autospec, sentinel, call
from datetime import datetime as dt

import numpy as np
import pandas as pd
import pytest
from dateutil.rrule import rrule, DAILY
from mockextras import when
from pandas.testing import assert_frame_equal

from arctic.date import DateRange, mktz
from arctic.exceptions import OverlappingDataException
from arctic.exceptions import UnhandledDtypeException
from arctic.tickstore.tickstore import TickStore
from arctic.tickstore.toplevel import TopLevelTickStore, TickStoreLibrary

utc = mktz('UTC')


def test_raise_exception_if_daterange_is_not_provided():
    store = TopLevelTickStore(Mock())
    with pytest.raises(Exception) as e:
        store._get_library_metadata(None)
    assert "A date range must be provided" in str(e.value)


def test_raise_exception_if_date_range_does_not_contain_start_date():
    store = TopLevelTickStore(Mock())
    dr = DateRange(start=None, end=dt(2011, 1, 1))
    with pytest.raises(Exception) as e:
        store._get_library_metadata(dr)
    assert "The date range {0} must contain a start and end date".format(dr) in str(e.value)


def test_raise_exception_if_date_range_does_not_contain_end_date():
    store = TopLevelTickStore(Mock())
    dr = DateRange(start=dt(2011, 1, 1), end=None)
    with pytest.raises(Exception) as e:
        store._get_library_metadata(dr)
    assert "The date range {0} must contain a start and end date".format(dr) in str(e.value)


def test_raise_exception_if_date_range_does_not_contain_start_and_end_date():
    store = TopLevelTickStore(Mock())
    dr = DateRange(start=None, end=None)
    with pytest.raises(Exception) as e:
        store._get_library_metadata(dr)
    assert "The date range {0} must contain a start and end date".format(dr) in str(e.value)


def test_raise_exception_and_log_an_error_if_an_invalid_library_name_is_added():
    arctic_lib = MagicMock()
    arctic_lib.arctic.__getitem__.side_effect = Exception()
    store = TopLevelTickStore(arctic_lib)
    with patch("arctic.tickstore.toplevel.logger") as mock_logger:
        with pytest.raises(Exception):
            store.add(None, "blah")
    mock_logger.error.assert_called_once_with("Could not load library")


def test_raise_exception_if_date_range_overlaps():
    self = create_autospec(TopLevelTickStore, _arctic_lib=MagicMock())
    self._get_library_metadata.return_value = [TickStoreLibrary('lib1', None), ]
    with pytest.raises(OverlappingDataException) as e:
        TopLevelTickStore.add(self, DateRange(start=dt(2010, 1, 1), end=dt(2011, 1, 1, 23, 59, 59, 999000)), "blah")
    assert "There are libraries that overlap with the date range:" in str(e.value)


@pytest.mark.parametrize(('start', 'end', 'expected_start', 'expected_end'),
                         [(dt(2010, 1, 1, tzinfo=mktz('UTC')), dt(2010, 12, 31, 23, 59, 59, 999000, tzinfo=mktz('UTC')),
                           dt(2010, 1, 1, tzinfo=mktz('UTC')), dt(2010, 12, 31, 23, 59, 59, 999000, tzinfo=mktz('UTC'))),
                          (dt(2010, 1, 1), dt(2010, 12, 31, 23, 59, 59, 999000), dt(2010, 1, 1, tzinfo=mktz('UTC')),
                           dt(2010, 12, 31, 23, 59, 59, 999000, tzinfo=mktz('UTC'))),
                          (dt(2009, 12, 31, 19, tzinfo=mktz('America/New_York')), dt(2010, 12, 31, 18, 59, 59, 999000, tzinfo=mktz('America/New_York')),
                           dt(2010, 1, 1, tzinfo=mktz('UTC')), dt(2010, 12, 31, 23, 59, 59, 999000, tzinfo=mktz('UTC')))
                          ])
def test_add_library_to_colllection_if_date_range_is_on_UTC_or_naive_day_boundaries(start, end, expected_start, expected_end):
    self = create_autospec(TopLevelTickStore, _arctic_lib=MagicMock(), _collection=MagicMock())
    self._get_library_metadata.return_value = []
    TopLevelTickStore.add(self, DateRange(start=start, end=end), "blah")
    self._collection.update_one.assert_called_once_with({'library_name': "blah"},
                                                        {'$set':
                                                         {'start': expected_start,
                                                          'end': expected_end}}, upsert=True)


@pytest.mark.parametrize(('start', 'end'),
                         [(dt(2010, 1, 1, 2, tzinfo=mktz('UTC')), dt(2011, 1, 1, tzinfo=mktz('UTC'))),
                          (dt(2010, 1, 1, tzinfo=mktz('UTC')), dt(2011, 1, 1, 2, tzinfo=mktz('UTC'))),
                          (dt(2010, 1, 1, 2, tzinfo=mktz('UTC')), dt(2011, 1, 1, 2, tzinfo=mktz('UTC'))),
                          (dt(2010, 1, 1, 2), dt(2011, 1, 1)),
                          (dt(2010, 1, 1), dt(2011, 1, 1, 2)),
                          (dt(2010, 1, 1, 2), dt(2011, 1, 1, 2)),
                          (dt(2009, 12, 31, 21, 10, tzinfo=mktz('America/New_York')), dt(2010, 12, 31, tzinfo=mktz('America/New_York'))),
                          (dt(2009, 12, 31, tzinfo=mktz('America/New_York')), dt(2010, 12, 31, tzinfo=mktz('America/New_York'))),
                          (dt(2009, 12, 31, 21, 10, tzinfo=mktz('America/New_York')), dt(2010, 12, 31, 9, 21, tzinfo=mktz('America/New_York')))
                          ])
def test_raise_error_add_library_is_called_with_a_date_range_not_on_day_boundaries(start, end):
    with pytest.raises(AssertionError) as e:
        self = create_autospec(TopLevelTickStore, _arctic_lib=MagicMock(), _collection=MagicMock())
        self._get_library_metadata.return_value = []
        TopLevelTickStore.add(self, DateRange(start=start, end=end), "blah")
    assert "Date range should fall on UTC day boundaries" in str(e.value)


@pytest.mark.parametrize(('start', 'end', 'expected_start_index', 'expected_end_index'),
                         [(dt(2010, 1, 1), dt(2010, 1, 5), 0, 3),
                          (dt(2010, 1, 1), dt(2010, 1, 6), 0, 3),
                          (dt(2010, 1, 1, 1), dt(2010, 1, 6), 1, 3),
                          (dt(2010, 1, 1, 1), dt(2010, 1, 4, 2), 1, 2),
                          (dt(2009, 1, 1), dt(2010, 1, 5), 0, 3),
                          ])
def test_slice_pandas_dataframe(start, end, expected_start_index, expected_end_index):
    top_level_tick_store = TopLevelTickStore(Mock())
    dates = pd.date_range('20100101', periods=5, freq='2D')
    data = pd.DataFrame(np.random.randn(5, 4), index=dates, columns=list('ABCD'))
    expected = data.iloc[expected_start_index:expected_end_index]
    result = top_level_tick_store._slice(data, start, end)
    assert_frame_equal(expected, result), '{}\n{}'.format(expected, result)


@pytest.mark.parametrize(('start', 'end', 'expected_start_index', 'expected_end_index'),
                         [(dt(2010, 1, 1, tzinfo=utc), dt(2010, 1, 5, tzinfo=utc), 0, 3),
                          (dt(2010, 1, 1, tzinfo=utc), dt(2010, 1, 6, tzinfo=utc), 0, 3),
                          (dt(2010, 1, 1, 1, tzinfo=utc), dt(2010, 1, 6, tzinfo=utc), 1, 3),
                          (dt(2010, 1, 1, 1, tzinfo=utc), dt(2010, 1, 4, 2, tzinfo=utc), 1, 2),
                          (dt(2009, 1, 1, tzinfo=utc), dt(2010, 1, 5, tzinfo=utc), 0, 3),
                          ])
def test_slice_list_of_dicts(start, end, expected_start_index, expected_end_index):
    top_level_tick_store = TopLevelTickStore(Mock())
    dates = list(rrule(DAILY, count=5, dtstart=dt(2010, 1, 1, tzinfo=utc), interval=2))
    data = [{'index': date, 'A': val} for date, val in zip(dates, range(5))]
    expected = data[expected_start_index:expected_end_index]
    result = top_level_tick_store._slice(data, start, end)
    assert expected == result


def test_write_pandas_data_to_right_libraries():
    self = create_autospec(TopLevelTickStore, _arctic_lib=MagicMock(), _collection=MagicMock())
    self._collection.find.return_value = [{'library_name': sentinel.libname1, 'start': sentinel.st1, 'end': sentinel.end1},
                                          {'library_name': sentinel.libname2, 'start': sentinel.st2, 'end': sentinel.end2}]
    slice1 = range(2)
    slice2 = range(4)
    when(self._slice).called_with(sentinel.data, sentinel.st1, sentinel.end1).then(slice1)
    when(self._slice).called_with(sentinel.data, sentinel.st2, sentinel.end2).then(slice2)
    mock_lib1 = Mock()
    mock_lib2 = Mock()
    when(self._arctic_lib.arctic.__getitem__).called_with(sentinel.libname1).then(mock_lib1)
    when(self._arctic_lib.arctic.__getitem__).called_with(sentinel.libname2).then(mock_lib2)
    with patch("arctic.tickstore.toplevel.to_dt") as patch_to_dt:
        patch_to_dt.side_effect = [sentinel.st1, sentinel.end1, sentinel.st2, sentinel.end2]
        TopLevelTickStore.write(self, 'blah', sentinel.data)
    mock_lib1.write.assert_called_once_with('blah', slice1)
    mock_lib2.write.assert_called_once_with('blah', slice2)


def test_read():
    self = create_autospec(TopLevelTickStore)
    tsl = TickStoreLibrary(create_autospec(TickStore), create_autospec(DateRange))
    self._get_libraries.return_value = [tsl, tsl]
    dr = create_autospec(DateRange)
    with patch('pandas.concat') as concat:
        res = TopLevelTickStore.read(self, sentinel.symbol, dr,
                                     columns=sentinel.include_columns,
                                     include_images=sentinel.include_images)
    assert concat.call_args_list == [call([tsl.library.read.return_value,
                                           tsl.library.read.return_value])]
    assert res == concat.return_value
    assert tsl.library.read.call_args_list == [call(sentinel.symbol, tsl.date_range.intersection.return_value,
                                                    sentinel.include_columns, include_images=sentinel.include_images),
                                               call(sentinel.symbol, tsl.date_range.intersection.return_value,
                                                    sentinel.include_columns, include_images=sentinel.include_images)]


def test_slice_raises():
    m = TopLevelTickStore(Mock())
    with pytest.raises(UnhandledDtypeException) as e:
        m._slice("abc", 1, 2)
    assert("Can't persist type" in str(e.value))



import time

import arctic.serialization.numpy_records as anr
from tests.unit.serialization.serialization_test_data import _mixed_test_data as input_test_data

df_serializer = anr.DataFrameSerializer()


def _bench(rounds, input_df, fast):
    fast = bool(fast)
    anr.set_fast_check_df_serializable(fast)
    start = time.time()
    for i in range(rounds):
        df_serializer.can_convert_to_records_without_objects(input_df, 'symA')
    print("Time per iteration (fast={}): {}".format(fast, (time.time() - start)/rounds))


# Results suggest significant speed improvements for
#   (1) large df with objects
#       Time per iteration (fast=False): 0.0281402397156
#       Time per iteration (fast=True):  0.00866063833237
#   (2) large multi-column df
#       Time per iteration (fast=False): 0.00556221961975
#       Time per iteration (fast=True):  0.00276621818542
#   (3) large multi-index df
#       Time per iteration (fast=False): 0.00640722036362
#       Time per iteration (fast=True):  0.00154552936554
def assess_speed(df_kind):
    rounds = 100
    input_df = input_test_data()[df_kind][0]
    orig_config = anr.FAST_CHECK_DF_SERIALIZABLE
    try:
        _bench(rounds, input_df, fast=False)

        _bench(rounds, input_df, fast=True)
    finally:
        anr.FAST_CHECK_DF_SERIALIZABLE = orig_config


def main():
    for df_kind in ('large_with_some_objects', 'large_multi_index', 'large_multi_column'):
        assess_speed(df_kind)


if __name__ == '__main__':
    main()



import time
from datetime import datetime as dt

import pytest
from mock import patch, MagicMock
from pandas.testing import assert_frame_equal
from pymongo.errors import OperationFailure

from arctic.arctic import Arctic, VERSION_STORE
from arctic.exceptions import LibraryNotFoundException, QuotaExceededException
from ..util import get_large_ts


def test_connect_to_Arctic_string(mongo_host):
    arctic = Arctic(mongo_host=mongo_host)
    assert arctic.list_libraries() == []
    assert arctic.mongo_host == mongo_host


def test_connect_to_Arctic_connection(mongo_server, mongo_host):
    arctic = Arctic(mongo_server.api)
    assert arctic.list_libraries() == []
    assert arctic.mongo_host == mongo_host


def test_reset_Arctic(mongo_host, library_name):
    arctic = Arctic(mongo_host=mongo_host)
    arctic.list_libraries()
    arctic.initialize_library(library_name, VERSION_STORE)
    c = arctic._conn
    assert arctic[library_name]._arctic_lib._curr_conn is c
    arctic.reset()
    assert c is not arctic._conn
    assert len(c.nodes) == 0
    assert arctic[library_name]._arctic_lib._curr_conn is arctic._conn


def test_re_authenticate_on_arctic_reset(mongo_host, library_name):
    from collections import namedtuple
    Cred = namedtuple('Cred', 'user, password')
    with patch('arctic.arctic.authenticate') as auth_mock, \
            patch('arctic.arctic.get_auth') as get_auth_mock:
        auth_mock.return_value = True
        get_auth_mock.return_value = Cred(user='a_username', password='a_passwd')
        arctic = Arctic(mongo_host=mongo_host)
        arctic.initialize_library(library_name, VERSION_STORE)
        vstore = arctic[library_name]
        vstore.list_symbols()
        auth_mock.reset_mock()
        arctic.reset()
        assert auth_mock.call_count > 0
        auth_mock.reset_mock()
        vstore.list_symbols()
        assert auth_mock.call_count == 0


def test_simple(library):
    sym = 'symbol'
    data = get_large_ts(100)

    library.write(sym, data)
    orig = dt.now()
    time.sleep(1)  # Move the timestamp on 1ms
    data2 = get_large_ts(100)
    library.write(sym, data2, prune_previous_version=False)

    # Get the timeseries, it should be the same
    read2 = library.read(sym).data
    assert_frame_equal(read2, data2)

    # Ensure we can get the previous version
    read = library.read(sym, as_of=orig).data
    assert_frame_equal(read, data)


def test_indexes(arctic):
    def makeMongo44Index(oldindex):
        # make a <=mongo42 index look like a >=mongo44 index
        fixed = dict()
        for k in sorted(oldindex.keys()):
            t2=dict(sorted(oldindex[k].items(), reverse=True))
            t2.pop('ns')
            fixed[k]=t2
        return fixed

    c = arctic._conn
    arctic.initialize_library("library", VERSION_STORE, segment='month')
    chunk = c.arctic.library.index_information()
    index_version = chunk['_id_']['v']  # Mongo 3.2 has index v1, 3.4 and 3.5 have v2 (3.4 can run in compabitility mode with v1)

    test_chunk = {u'_id_': {u'key': [(u'_id', 1)], u'ns': u'arctic.library', u'v': index_version},
                     u'symbol_1_parent_1_segment_1': {u'background': True,
                                                      u'key': [(u'symbol', 1),
                                                               (u'parent', 1),
                                                               (u'segment', 1)],
                                                      u'ns': u'arctic.library',
                                                      u'unique': True,
                                                      u'v': index_version},
                     u'symbol_1_sha_1': {u'background': True,
                                         u'key': [(u'symbol', 1), (u'sha', 1)],
                                         u'ns': u'arctic.library',
                                         u'unique': True,
                                         u'v': index_version},
                     u'symbol_hashed': {u'background': True,
                                        u'key': [(u'symbol', u'hashed')],
                                        u'ns': u'arctic.library',
                                        u'v': index_version},
                     u'symbol_1_sha_1_segment_1': {u'background': True,
                                                   u'key': [(u'symbol', 1), (u'sha', 1), (u'segment', 1)],
                                                   u'ns': u'arctic.library',
                                                   u'unique': True,
                                                   u'v': index_version}}

    test_chunk_44 = makeMongo44Index(test_chunk)
    assert chunk == test_chunk or chunk == test_chunk_44

    snapshots = c.arctic.library.snapshots.index_information()
    test_snapshots = {u'_id_': {u'key': [(u'_id', 1)],
                                               u'ns': u'arctic.library.snapshots',
                                               u'v': index_version},
                                     u'name_1': {u'background': True,
                                                 u'key': [(u'name', 1)],
                                                 u'ns': u'arctic.library.snapshots',
                                                 u'unique': True,
                                                 u'v': index_version}}
    test_snapshots_44 = makeMongo44Index(test_snapshots)
    assert snapshots == test_snapshots or snapshots == test_snapshots_44

    versions = c.arctic.library.versions.index_information()
    test_versions = {'_id_': {'v': index_version, 'key': [('_id', 1)],
                              'ns': 'arctic.library.versions'},
                     'symbol_1__id_-1': {'v': index_version, 'key': [('symbol', 1), ('_id', -1)],
                                         'ns': 'arctic.library.versions', 'background': True},
                     'symbol_1_version_-1': {'v': index_version, 'unique': True, 'key': [('symbol', 1), ('version', -1)],
                                             'ns': 'arctic.library.versions', 'background': True},
                     'versionstore_idx': {'v': index_version, 'key': [('symbol', 1), ('version', -1), ('metadata.deleted', 1)],
                                          'ns': 'arctic.library.versions', 'background': True},
                     'parent_1': {'v': index_version, 'key': [('parent', 1)],
                                            'ns': 'arctic.library.versions', 'background': True}}
    test_versions_44 = makeMongo44Index(test_versions)
    assert versions == test_versions or versions == test_versions_44

    version_nums = c.arctic.library.version_nums.index_information()
    test_version_nums = {u'_id_': {u'key': [(u'_id', 1)],
                                               u'ns': u'arctic.library.version_nums',
                                               u'v': index_version},
                                     u'symbol_1': {u'background': True,
                                                   u'key': [(u'symbol', 1)],
                                                   u'ns': u'arctic.library.version_nums',
                                                   u'unique': True,
                                                   u'v': index_version}}
    test_version_nums_44 = makeMongo44Index(test_version_nums)
    assert version_nums == test_version_nums or version_nums == test_version_nums_44


def test_delete_library(arctic, library, library_name):
    mongo = arctic._conn
    # create a library2 library too - ensure that this isn't deleted
    arctic.initialize_library('user.library2', VERSION_STORE, segment='month')
    library.write('asdf', get_large_ts(1))
    assert 'TEST' in mongo.arctic_test.list_collection_names()
    assert 'TEST.versions' in mongo.arctic_test.list_collection_names()
    assert 'library2' in mongo.arctic_user.list_collection_names()
    assert 'library2.versions' in mongo.arctic_user.list_collection_names()

    arctic.delete_library(library_name)
    assert 'TEST' not in mongo.arctic_user.list_collection_names()
    assert 'TEST.versions' not in mongo.arctic_user.list_collection_names()
    with pytest.raises(LibraryNotFoundException):
        arctic[library_name]
    with pytest.raises(LibraryNotFoundException):
        arctic['arctic_{}'.format(library_name)]
    assert 'library2' in mongo.arctic_user.list_collection_names()
    assert 'library2.versions' in mongo.arctic_user.list_collection_names()


def test_quota(arctic, library, library_name):
    thing = list(range(100))
    library._arctic_lib.set_quota(10)
    assert arctic.get_quota(library_name) == 10
    assert library._arctic_lib.get_quota() == 10
    library.write('thing', thing)
    with pytest.raises(QuotaExceededException):
        library.write('ts', thing)
        library.write('ts', thing)
        library.write('ts', thing)
        library.write('ts', thing)
    with pytest.raises(QuotaExceededException):
        arctic.check_quota(library_name)


def test_check_quota(arctic, library, library_name):
    with patch('arctic.arctic.logger.info') as info:
        arctic.check_quota(library_name)
    assert info.call_count == 1


def test_default_mongo_retry_timout():
    now = time.time()
    with pytest.raises(LibraryNotFoundException):
        Arctic('unresolved-host', serverSelectionTimeoutMS=0)['some.lib']
    assert time.time() - now < 1.


def test_lib_rename(arctic):
    arctic.initialize_library('test')
    l = arctic['test']
    l.write('test_data', 'abc')
    arctic.rename_library('test', 'new_name')
    l = arctic['new_name']
    assert(l.read('test_data').data == 'abc')
    with pytest.raises(LibraryNotFoundException) as e:
        l = arctic['test']
    assert('Library test' in str(e.value))
    assert('test' not in arctic.list_libraries())


def test_lib_rename_namespace(arctic):
    arctic.initialize_library('namespace.test')
    l = arctic['namespace.test']
    l.write('test_data', 'abc')

    with pytest.raises(ValueError) as e:
        arctic.rename_library('namespace.test', 'new_namespace.test')
    assert('Collection can only be renamed in the same database' in str(e.value))

    arctic.rename_library('namespace.test', 'namespace.newlib')
    l = arctic['namespace.newlib']
    assert(l.read('test_data').data == 'abc')

    with pytest.raises(LibraryNotFoundException) as e:
        l = arctic['namespace.test']
    assert('Library namespace.test' in str(e.value))
    assert('namespace.test' not in arctic.list_libraries())


def test_lib_type(arctic):
    arctic.initialize_library('test')
    assert(arctic.get_library_type('test') == VERSION_STORE)


def test_library_exists(arctic):
    arctic.initialize_library('test')
    assert arctic.library_exists('test')
    assert not arctic.library_exists('nonexistentlib')


def test_library_exists_no_auth(arctic):
    arctic.initialize_library('test')
    with patch('arctic.arctic.ArcticLibraryBinding') as AB:
        AB.return_value = MagicMock(
            get_library_type=MagicMock(side_effect=OperationFailure("not authorized on arctic to execute command")))
        assert arctic.library_exists('test')
        assert AB.return_value.get_library_type.called
        assert not arctic.library_exists('nonexistentlib')


def test_list_libraries_cached(arctic):
    # default in arctic is to cache list_libraries.
    libs = ['test1', 'test2']
    for lib in libs:
        arctic.initialize_library(lib)

    # Cached data should have been appended to cache.
    assert sorted(libs) == sorted(arctic.list_libraries()) == sorted(arctic._list_libraries())

    # Should default to uncached list_libraries if cache is empty.
    with patch('arctic.arctic.Arctic._list_libraries', return_value=libs) as uncached_list_libraries:
        # Empty cache manually.
        arctic._conn.meta_db.cache.remove({})
        assert arctic._list_libraries_cached() == libs
        uncached_list_libraries.assert_called()

    # Reload cache and check that it has data
    arctic.reload_cache()
    assert sorted(arctic._cache.get('list_libraries')) == sorted(libs)

    # Should fetch it from cache the second time.
    with patch('arctic.arctic.Arctic._list_libraries', return_value=libs) as uncached_list_libraries:
        assert sorted(arctic._list_libraries_cached()) == sorted(libs)
        uncached_list_libraries.assert_not_called()


def test_initialize_library_adds_to_cache(arctic):
    libs = ['test1', 'test2']

    for lib in libs:
        arctic.initialize_library(lib)

    arctic.reload_cache()
    assert arctic._list_libraries_cached() == arctic._list_libraries()

    # Add another lib
    arctic.initialize_library('test3')

    assert sorted(arctic._cache.get('list_libraries')) == ['test1', 'test2', 'test3']


def test_cache_does_not_return_stale_data(arctic):
    libs = ['test1', 'test2']

    for lib in libs:
        arctic.initialize_library(lib)

    arctic.reload_cache()
    assert arctic._list_libraries_cached() == arctic._list_libraries()

    time.sleep(0.2)

    # Should call uncached list_libraries if the data is stale according to caller.
    with patch('arctic.arctic.Arctic._list_libraries', return_value=libs) as uncached_list_libraries:
        assert arctic._list_libraries_cached(newer_than_secs=0.1) == libs
        uncached_list_libraries.assert_called()


def test_renaming_returns_new_name_in_cache(arctic):
    libs = ['test1', 'test2']

    for lib in libs:
        arctic.initialize_library(lib)

    assert sorted(arctic._list_libraries_cached()) == sorted(arctic._list_libraries())

    arctic.rename_library('test1', 'test3')

    assert sorted(arctic._list_libraries_cached()) == sorted(['test2', 'test3'])


def test_deleting_library_removes_it_from_cache(arctic):
    libs = ['test1', 'test2']

    for lib in libs:
        arctic.initialize_library(lib)

    arctic.delete_library('test1')

    assert arctic._list_libraries_cached() == arctic._list_libraries() == arctic.list_libraries() == ['test2']


def test_disable_cache_by_settings(arctic):
    lib = 'test1'
    arctic.initialize_library(lib)

    # Should be enabled by default
    assert arctic._list_libraries_cached() == arctic._list_libraries()

    arctic._cache.set_caching_state(enabled=False)

    # Should not return cached results now.
    with patch('arctic.arctic.Arctic._list_libraries', return_value=[lib]) as uncached_list_libraries:
        with patch('arctic.arctic.Arctic._list_libraries_cached', return_value=[lib]) as cached_list_libraries:
            arctic.list_libraries()
            uncached_list_libraries.assert_called()
            cached_list_libraries.assert_not_called()

    arctic._cache.set_caching_state(enabled=True)

    # Should used cached data again.
    with patch('arctic.arctic.Arctic._list_libraries', return_value=[lib]) as uncached_list_libraries_e:
        with patch('arctic.arctic.Arctic._list_libraries_cached', return_value=[lib]) as cached_list_libraries_e:
            arctic.list_libraries()
            uncached_list_libraries_e.assert_not_called()
            cached_list_libraries_e.assert_called()



from contextlib import contextmanager


@contextmanager
def enable_profiling_for_library(library):
    library._arctic_lib._db.set_profiling_level(2)
    yield library._arctic_lib._db['system.profile']
    library._arctic_lib._db.set_profiling_level(0, slow_ms=100)


import datetime
import re

import dateutil
import numpy as np
import pandas as pd
import pytz
import pytest
from mock import patch, Mock, sentinel
from numpy.testing import assert_array_equal
from pandas import Timestamp

import arctic.serialization.numpy_records as anr


class FastCheckSerializable(object):
    def __init__(self, enabled):
        self.enabled = bool(enabled)
        self.orig_setting = None

    def __enter__(self):
        self.orig_setting = anr.FAST_CHECK_DF_SERIALIZABLE
        anr.set_fast_check_df_serializable(self.enabled)

    def __exit__(self, *args):
        anr.set_fast_check_df_serializable(self.orig_setting)


def test_to_primitive_timestamps():
    arr = anr._to_primitive(np.array([Timestamp('2010-11-12 00:00:00')]))
    assert_array_equal(arr, np.array([Timestamp('2010-11-12 00:00:00').value], dtype='datetime64[ns]'))


def test_to_primitive_fixed_length_strings():
    mydf = pd.DataFrame({'a': ['abc', u'xyz', '']})
    primitives_arr = anr._to_primitive(np.array(mydf.a.values), string_max_len=32)
    assert_array_equal(primitives_arr, np.array([u'abc', u'xyz', u''], dtype='U32'))
    assert primitives_arr.dtype == np.dtype('U32')


@pytest.mark.parametrize("fast_serializable_check", (True, False))
def test_can_convert_to_records_without_objects_returns_false_on_exception_in_to_records(fast_serializable_check):
    with FastCheckSerializable(fast_serializable_check):
        store = anr.PandasSerializer()
        mymock = Mock(side_effect=TypeError('uhoh'))
        if fast_serializable_check:
            store.fast_check_serializable = mymock
        else:
            store._to_records = mymock
        with patch('arctic.serialization.numpy_records.log') as mock_log:
            assert store.can_convert_to_records_without_objects(sentinel.df, 'my_symbol') is False

        assert 'Pandas dataframe my_symbol caused exception' in str(mock_log.warning.call_args)
        if fast_serializable_check:
            store.fast_check_serializable.assert_called_once_with(sentinel.df)
        else:
            store._to_records.assert_called_once_with(sentinel.df)


@pytest.mark.parametrize("fast_serializable_check", (True, False))
def test_can_convert_to_records_without_objects_returns_false_when_records_have_object_dtype(fast_serializable_check):
    with FastCheckSerializable(fast_serializable_check):
        store = anr.PandasSerializer()
        mymock = Mock(return_value=(np.array(['a', 'b', None, 'd']), None))
        if fast_serializable_check:
            store.fast_check_serializable = mymock
        else:
            store._to_records = mymock
        with patch('arctic.serialization.numpy_records.log') as mock_log:
            assert store.can_convert_to_records_without_objects(sentinel.df, 'my_symbol') is False
        mock_log.warning.assert_called_once_with('Pandas dataframe my_symbol contains Objects, saving as Blob')
        if fast_serializable_check:
            store.fast_check_serializable.assert_called_once_with(sentinel.df)
        else:
            store._to_records.assert_called_once_with(sentinel.df)


@pytest.mark.parametrize("fast_serializable_check", (True, False))
def test_can_convert_to_records_without_objects_returns_false_when_records_have_arrays_in_them(fast_serializable_check):
    with FastCheckSerializable(fast_serializable_check):
        store = anr.PandasSerializer()
        mymock = Mock(return_value=(np.rec.array([(1356998400000000000, ['A', 'BC'])], dtype=[('index', '<M8[ns]'), ('values', 'S2', (2,))]), None))
        if fast_serializable_check:
            store.fast_check_serializable = mymock
        else:
            store._to_records = mymock
        with patch('arctic.serialization.numpy_records.log') as mock_log:
            assert store.can_convert_to_records_without_objects(sentinel.df, 'my_symbol') is False
        mock_log.warning.assert_called_once_with('Pandas dataframe my_symbol contains >1 dimensional arrays, saving as Blob')
        if fast_serializable_check:
            store.fast_check_serializable.assert_called_once_with(sentinel.df)
        else:
            store._to_records.assert_called_once_with(sentinel.df)


@pytest.mark.parametrize("fast_serializable_check", (True, False))
def test_can_convert_to_records_without_objects_returns_true_otherwise(fast_serializable_check):
    with FastCheckSerializable(fast_serializable_check):
        store = anr.PandasSerializer()
        mymock = Mock(return_value=(np.rec.array([(1356998400000000000, 'a')], dtype=[('index', '<M8[ns]'), ('values', 'S2')]), None))
        if fast_serializable_check:
            store.fast_check_serializable = mymock
        else:
            store._to_records = mymock
        with patch('arctic.serialization.numpy_records.log') as mock_log:
            assert store.can_convert_to_records_without_objects(sentinel.df, 'my_symbol') is True
        assert mock_log.warning.call_count == 0
        if fast_serializable_check:
            store.fast_check_serializable.assert_called_once_with(sentinel.df)
        else:
            store._to_records.assert_called_once_with(sentinel.df)


@pytest.mark.parametrize(
    ["tz", "expected_tz_str_pat"],
    [
        ("UTC", r"^UTC$"),
        (pytz.utc, r"^UTC$"),
        (dateutil.tz.tzutc(), r"^UTC$"),
        (dateutil.tz.gettz("UTC"), r"^dateutil/.+UTC"),
        *([(datetime.timezone.utc, r"^UTC$")] if hasattr(datetime, "timezone") else []),
        (pytz.timezone("Europe/London"), r"^Europe/London$"),
        (pytz.timezone("America/New_York"), r"^America/New_York$"),
        (dateutil.tz.gettz("Europe/London"), r"^dateutil/.+Europe/London"),
        (dateutil.tz.gettz("America/New_York"), r"^dateutil/.+America/New_York"),
    ],
)
@pytest.mark.parametrize("index_nlevels", [1, 2])
def test_dataframe_serializer_serialize_tz_index(
    tz, expected_tz_str_pat, index_nlevels
):
    if index_nlevels == 1:
        index = pd.date_range("2023-02-04", "2023-02-06", tz=tz, name="idx_lvl1")
    else:
        index = pd.MultiIndex.from_arrays(
            [
                pd.date_range(
                    "2023-02-04", "2023-02-06", tz=tz, name="idx_lvl{}".format(idx_lvl)
                )
                for idx_lvl in range(1, index_nlevels + 1)
            ]
        )

    df = pd.DataFrame({"a": [1, 2, 3]}, index=index)
    serializer = anr.DataFrameSerializer()

    result_records, result_dtype = serializer.serialize(df)

    expected_dtype_descr = [
        *(
            ("idx_lvl{}".format(idx_lvl), str(np.dtype("datetime64[ns]")))
            for idx_lvl in range(1, index_nlevels + 1)
        ),
        ("a", str(np.dtype("int64"))),
    ]
    expected_records = np.rec.fromarrays(
        [
            *(
                [
                    pd.date_range("2023-02-04", "2023-02-06", tz=tz)
                    .tz_convert("UTC")
                    .values
                ]
                * index_nlevels
            ),
            np.array([1, 2, 3], dtype=np.int64),
        ],
        dtype=np.dtype(expected_dtype_descr),
    )

    np.testing.assert_array_equal(result_records, expected_records)

    assert result_dtype.metadata["columns"] == ["a"]
    if index_nlevels == 1:
        assert result_dtype.metadata["index"] == ["idx_lvl1"]
        assert re.search(expected_tz_str_pat, result_dtype.metadata["index_tz"])
    else:
        assert result_dtype.metadata["index"] == [
            "idx_lvl{}".format(idx_lvl) for idx_lvl in range(1, index_nlevels + 1)
        ]
        for index_lvl_tz in result_dtype.metadata["index_tz"]:
            assert re.search(expected_tz_str_pat, index_lvl_tz)


@pytest.mark.parametrize("fast_serializable_check", (False, True))
def test_can_convert_to_records_mixed_object_column_string_nan(fast_serializable_check):
    with FastCheckSerializable(fast_serializable_check):
        serializer = anr.DataFrameSerializer()

        df = pd.DataFrame({'a': [1, 3, 4], 'b': [1.2, 8.0, 0.2]})
        assert serializer.can_convert_to_records_without_objects(df, 'my_symbol')

        df = pd.DataFrame({'a': [1, 3, 4], 'b': [1, 8.0, 2]})
        assert serializer.can_convert_to_records_without_objects(df, 'my_symbol')

        df = pd.DataFrame({'a': [1, 3, 4], 'b': [1.2, 8.0, np.NaN]})
        assert serializer.can_convert_to_records_without_objects(df, 'my_symbol')

        df = pd.DataFrame({'a': ['abc', 'cde', 'def'], 'b': [1.2, 8.0, np.NaN]})
        assert serializer.can_convert_to_records_without_objects(df, 'my_symbol')

        df = pd.DataFrame({'a': [u'abc', u'cde', 'def'], 'b': [1.2, 8.0, np.NaN]})
        assert serializer.can_convert_to_records_without_objects(df, 'my_symbol')

        df = pd.DataFrame({'a': [u'abc', u'cde', 'def'], 'b': [1.2, '8.0', np.NaN]})
        assert not serializer.can_convert_to_records_without_objects(df, 'my_symbol')

        # Do not serialize and force-stringify None
        df = pd.DataFrame({'a': ['abc', None, 'def'], 'b': [1.2, 8.0, np.NaN]})
        assert not serializer.can_convert_to_records_without_objects(df, 'my_symbol')

        # Do not serialize and force-stringify np.NaN among strings, rather pickle
        df = pd.DataFrame({'a': ['abc', np.NaN, 'def'], 'b': [1.2, 8.0, np.NaN]})
        assert not serializer.can_convert_to_records_without_objects(df, 'my_symbol')



""" The Arctic TimeSeries and Tick store."""

import pandas

from .arctic import Arctic, register_library_type
from .arctic import VERSION_STORE, TICK_STORE, CHUNK_STORE, METADATA_STORE, SEQUENCE_STORE
from .store._ndarray_store import NdarrayStore
from .store._pandas_ndarray_store import PandasDataFrameStore, PandasSeriesStore, PandasPanelStore
from .store.version_store import register_versioned_storage, register_version

try:
    from packaging.version import Version
    import importlib_metadata
    str_version = importlib_metadata.version("arctic_new")
    parsed_version = Version(str_version)
    num_version = sum([1000 ** i * v for i, v in enumerate(reversed(parsed_version.release))])
    register_version(str_version, num_version)
except Exception as e:
    __version__ = None
    __version_parts__ = tuple()
    __version_numerical__ = 0
else:
    __version__ = str_version
    __version_parts__ = parsed_version.release
    __version_numerical__ = num_version


register_versioned_storage(PandasDataFrameStore)
register_versioned_storage(PandasSeriesStore)
register_versioned_storage(NdarrayStore)




import time

from mock import patch, ANY, call

from arctic.auth import Credential
from arctic.scripts import arctic_prune_versions as mpv
from ...util import run_as_main


def test_prune_versions_symbol(mongo_host, library, library_name):
    with patch('arctic.scripts.arctic_prune_versions.prune_versions', autospec=True) as prune_versions, \
            patch('arctic.scripts.utils.get_auth', return_value=Credential('admin', 'adminuser', 'adminpwd')), \
            patch('pymongo.database.Database.authenticate', return_value=True):

        run_as_main(mpv.main, '--host', mongo_host, '--library', library_name, '--symbols', 'sym1,sym2')
        prune_versions.assert_has_calls([call(ANY, ['sym1', 'sym2'], 10)])


def test_prune_versions_full(mongo_host, library, library_name):
    with patch('arctic.scripts.arctic_prune_versions.do_db_auth', return_value=True):
        # Write some stuff with snapshots
        library.snapshot('snap')
        library.write('symbol', "val1")
        library.write('symbol', "val2")
        library.snapshot('snap1')
        library.write('symbol', "val3")

        # Prune older than 10 mins - nothing deleted
        run_as_main(mpv.main, '--host', mongo_host, '--library', library_name, '--keep-mins', 10)
        assert [x['version'] for x in library.list_versions('symbol')] == [3, 2, 1]
        # Prune older than 0 minutes, v1 deleted
        run_as_main(mpv.main, '--host', mongo_host, '--library', library_name, '--keep-mins', 0)
        assert [x['version'] for x in library.list_versions('symbol')] == [3, 2]

        # Delete the snapshots
        library.delete_snapshot('snap')
        library.delete_snapshot('snap1')
        run_as_main(mpv.main, '--host', mongo_host, '--library', library_name, '--keep-mins', 0)
        assert [x['version'] for x in library.list_versions('symbol')] == [3]


def test_keep_recent_snapshots(library):
    library.write("cherry", "blob")
    half_a_day_ago = time.time() - (3600 * 12.)
    with patch('time.time', return_value=half_a_day_ago):
        library.snapshot("snappy")
    library._snapshots.delete_one({"name": "snappy"})

    mpv.prune_versions(library, ["cherry"], 10)

    assert len(library._versions.find_one({"symbol": "cherry"}).get("parent", [])) == 1


def test_fix_broken_snapshot_references(library):
    library.write("cherry", "blob")
    one_day_ago = time.time() - (3600 * 24.) - 10  # make sure we are a few seconds before 24 hours
    with patch('time.time', return_value=one_day_ago):
        library.snapshot("snappy")
    library._snapshots.delete_one({"name": "snappy"})

    mpv.prune_versions(library, ["cherry"], 10)

    assert library._versions.find_one({"symbol": "cherry"}).get("parent", []) == []


def test_keep_only_one_version(library):
    library.write("cherry", "blob")
    library.write("cherry", "blob")
    one_day_ago = time.time() - (3600 * 24.) - 10  # make sure we are a few seconds before 24 hours
    with patch('time.time', return_value=one_day_ago):
        library.snapshot("snappy")
    library._snapshots.delete_one({"name": "snappy"})

    mpv.prune_versions(library, ["cherry"], 0)

    assert len(list(library._versions.find({"symbol": "cherry"}))) == 1
